from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import onnx
import unittest

from tests._test_utils import _onnx_create_model
from onnx import helper, numpy_helper, ModelProto, TensorProto
from onnx_coreml import convert
from coremltools.proto import NeuralNetwork_pb2 #type: ignore

def _make_model_acos_exp_topk(): # type: (...) -> ModelProto
  '''
  make a very simple model for testing: input->clip->exp->topk->2 outputs
  '''
  inputs = [('input0', (10,), TensorProto.FLOAT), ('K', (1,), TensorProto.INT64)]
  outputs = [('output_values', (3,), TensorProto.FLOAT), ('output_indices', (3,), TensorProto.INT64)]
  acos = helper.make_node("Acos",
                          inputs=[inputs[0][0]],
                          outputs=['acos_out'])
  exp = helper.make_node("Exp",
                        inputs=[acos.output[0]],
                        outputs=['exp_out'])
  topk = helper.make_node("TopK",
                        inputs=[exp.output[0], inputs[1][0]],
                        outputs=[outputs[0][0], outputs[1][0]],
                        axis=0)
  return _onnx_create_model([acos, exp, topk], inputs, outputs)

def _make_model_flatten_axis3(): # type: (...) -> ModelProto
  '''
  make a simple model: 4-D input -> flatten (axis=3)-> output
  '''
  inputs = [('input', (1,3,10,20), TensorProto.FLOAT)]
  outputs = [('output', (30,20), TensorProto.FLOAT)]
  flatten = helper.make_node("Flatten",
                            inputs=[inputs[0][0]],
                            outputs=[outputs[0][0]],
                            axis=3)
  return _onnx_create_model([flatten], inputs, outputs)


class CustomLayerTest(unittest.TestCase):

  def test_unsupported_ops(self):  # type: () -> None

    onnx_model = _make_model_acos_exp_topk()
    coreml_model = convert(onnx_model, add_custom_layers=True)

    spec = coreml_model.get_spec()
    layers = spec.neuralNetwork.layers
    self.assertIsNotNone(layers[0].custom)
    self.assertIsNotNone(layers[2].custom)
    self.assertEqual('Acos', layers[0].custom.className)
    self.assertEqual('TopK', layers[2].custom.className)

  def test_unsupported_ops_provide_functions(self):  # type: () -> None

    def convert_acos(builder, node, graph, err):
      params = NeuralNetwork_pb2.CustomLayerParams()
      params.className = node.op_type
      params.description = "Custom layer that corresponds to the ONNX op {}".format(node.op_type, )

      builder.add_custom(
        name=node.name,
        input_names=node.inputs,
        output_names=node.outputs,
        custom_proto_spec=params
      )

    def convert_topk(builder, node, graph, err):
      params = NeuralNetwork_pb2.CustomLayerParams()
      params.className = node.op_type
      params.description = "Custom layer that corresponds to the ONNX op {}".format(node.op_type, )
      params.parameters["axis"].intValue = node.attrs.get('axis', -1)

      builder.add_custom(
        name=node.name,
        input_names=node.inputs,
        output_names=node.outputs,
        custom_proto_spec=params
      )

    onnx_model = _make_model_acos_exp_topk()
    coreml_model = convert(model=onnx_model,
                           add_custom_layers=True,
                           custom_conversion_functions={'Acos':convert_acos, 'TopK':convert_topk})

    spec = coreml_model.get_spec()
    layers = spec.neuralNetwork.layers
    self.assertIsNotNone(layers[0].custom)
    self.assertIsNotNone(layers[2].custom)
    self.assertEqual('Acos', layers[0].custom.className)
    self.assertEqual('TopK', layers[2].custom.className)
    self.assertEqual(0, layers[2].custom.parameters['axis'].intValue)

  def test_node_name_type_custom_functions(self):  # type: () -> None
    def convert_acos(builder, node, graph, err):
      params = NeuralNetwork_pb2.CustomLayerParams()
      params.className = node.op_type
      params.description = "Custom layer that corresponds to the ONNX op {}".format(node.op_type, )

      builder.add_custom(
        name=node.name,
        input_names=node.inputs,
        output_names=node.outputs,
        custom_proto_spec=params
      )

    def convert_topk_generic(builder, node, graph, err):
      params = NeuralNetwork_pb2.CustomLayerParams()
      params.className = node.op_type
      params.description = "Custom layer that corresponds to the ONNX op {}".format(node.op_type, )
      params.parameters["axis"].intValue = node.attrs.get('axis', -1)
      params.parameters["k"].intValue = node.attrs['k']

      builder.add_custom(
        name=node.name,
        input_names=node.inputs,
        output_names=node.outputs,
        custom_proto_spec=params
      )

    def convert_topk_node_specific(builder, node, graph, err):
      params = NeuralNetwork_pb2.CustomLayerParams()
      params.className = node.op_type
      params.description = "Custom layer that corresponds to the ONNX op {}".format(node.op_type, )
      params.parameters["axis"].intValue = node.attrs.get('axis', -1)

      builder.add_custom(
        name=node.name,
        input_names=node.inputs,
        output_names=node.outputs,
        custom_proto_spec=params
      )

    onnx_model = _make_model_acos_exp_topk()
    onnx.save_model(onnx_model, 'acos.onnx')
    coreml_model = convert(model=onnx_model,
                           add_custom_layers=True,
                           custom_conversion_functions={'Acos':convert_acos, 'TopK':convert_topk_generic,
                                                        'output_values_output_indices':convert_topk_node_specific})

    spec = coreml_model.get_spec()
    layers = spec.neuralNetwork.layers
    self.assertIsNotNone(layers[0].custom)
    self.assertIsNotNone(layers[2].custom)
    self.assertEqual('Acos', layers[0].custom.className)
    self.assertEqual('TopK', layers[2].custom.className)
    self.assertEqual(0, layers[2].custom.parameters['axis'].intValue)
  
  def test_unsupported_op_attribute(self):  # type: () -> None
    onnx_model = _make_model_flatten_axis3()
    coreml_model = convert(onnx_model, add_custom_layers=True)

    spec = coreml_model.get_spec()
    layers = spec.neuralNetwork.layers
    self.assertIsNotNone(layers[0].custom)
    self.assertEqual('Flatten', layers[0].custom.className)

  def test_unsupported_op_attribute_provide_functions(self):  # type: () -> None

    def convert_flatten(builder, node, graph, err):
      params = NeuralNetwork_pb2.CustomLayerParams()
      params.className = node.op_type
      params.description = "Custom layer that corresponds to the ONNX op {}".format(node.op_type, )
      params.parameters["axis"].intValue = node.attrs['axis']
      
      builder.add_custom(
        name=node.name,
        input_names=node.inputs,
        output_names=node.outputs,
        custom_proto_spec=params
      )

    def test_conversion(onnx_model, add_custom_layers=False):
      coreml_model = convert(onnx_model, add_custom_layers=add_custom_layers,
                           custom_conversion_functions={'Flatten': convert_flatten})

      spec = coreml_model.get_spec()
      layers = spec.neuralNetwork.layers
      self.assertIsNotNone(layers[0].custom)
      self.assertEqual('Flatten', layers[0].custom.className)
      self.assertEqual(3, layers[0].custom.parameters['axis'].intValue)
  
    onnx_model = _make_model_flatten_axis3()
    # Test with add_custom_layers True
    convert(onnx_model, add_custom_layers=True,
                           custom_conversion_functions={'Flatten': convert_flatten})
    
    # Test with add_custom_layers False
    convert(onnx_model, add_custom_layers=False,
                           custom_conversion_functions={'Flatten': convert_flatten})

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import click
from onnx import onnx_pb
from onnx_coreml import convert
from typing import Text, IO


@click.command(
    help='convert ONNX model to CoreML model',
    context_settings={
        'help_option_names': ['-h', '--help']
    }
)
@click.argument('onnx_model', type=click.File('rb'))
@click.option('-o', '--output', required=True,
              type=str,
              help='Output path for the CoreML *.mlmodel file')
def onnx_to_coreml(onnx_model, output):  # type: (IO[str], str) -> None
    onnx_model_proto = onnx_pb.ModelProto()
    onnx_model_proto.ParseFromString(onnx_model.read())
    coreml_model = convert(onnx_model_proto)
    coreml_model.save(output)


if __name__ == '__main__':
    onnx_to_coreml()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .converter import convert

__all__ = ['convert']



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Dict, Text, Any, Callable
from coremltools.models.neural_network import NeuralNetworkBuilder  #type: ignore
from ._graph import Node, Graph

class ErrorHandling(object):
  '''
  To handle errors and addition of custom layers
  '''

  def __init__(self,
               add_custom_layers = False, # type: bool
               custom_conversion_functions = dict(), # type: Dict[Text, Any]
               custom_layer_nodes = [], # type : List[Node]
               disable_coreml_rank5_mapping = False
               ):
      # type: (...) -> None
      self.add_custom_layers = add_custom_layers
      self.custom_conversion_functions = custom_conversion_functions
      self.custom_layer_nodes = custom_layer_nodes
      self.disable_coreml_rank5_mapping = disable_coreml_rank5_mapping
      # TODO: Remove following error message once, disable_coreml_rank5_mapping is default to True
      self.coreml_3_rerun_message = ''
      if not disable_coreml_rank5_mapping:
        self.coreml_3_rerun_message = '\nPlease try converting again with disable_coreml_rank5_mapping=True' \
                                                        ' and coremltools 3.0 latest beta'


  def unsupported_op(self,
                     node,  # type: Node
                    ):
      # type: (...) -> Callable[[Any, Node, Graph, ErrorHandling], None]
      '''
      Either raise an error for an unsupported op type or return custom layer add function
      '''
      if self.add_custom_layers:
        from ._operators import _convert_custom
        return _convert_custom
      else:
        raise TypeError(
          "ONNX node of type {} is not supported. {}\n".format(node.op_type, self.coreml_3_rerun_message)
        )


  def unsupported_op_configuration(self,
                                   builder, # type: NeuralNetworkBuilder
                                   node, # type: Node
                                   graph, # type: Graph
                                   err_message, # type: Text
                                   ):
      # type: (...) -> None
      '''
      Either raise an error for an unsupported attribute or add a custom layer.
      '''
      if self.add_custom_layers:
        from ._operators import _convert_custom
        _convert_custom(builder, node, graph, self)
      else:
        raise TypeError(
          "Error while converting op of type: {}. Error message: {} {}\n".format(node.op_type, err_message,
                                                                                  self.coreml_3_rerun_message)
        )


  def missing_initializer(self,
                          node, # type: Node
                          err_message, # type: Text
                          ):
      # type: (...) -> None
      '''
      Missing initializer error
      '''
      raise ValueError(
        "Missing initializer error in op of type {}, with input name = {}, "
        "output name = {}. Error message: {} {}\n".
        format(node.op_type, node.inputs[0], node.outputs[0], err_message, self.coreml_3_rerun_message)
      )

  def unsupported_feature_warning(self,
                                  node, # type: Node
                                  err_message, # type: Text
                                  ):
      # type: (...) -> None
      '''
      Unsupported feature warning
      '''
      print(
        "Warning: Unsupported Feature in op of type {}, with input name = {}, "
        "output name = {}. Warning message: {} {}\n".
        format(node.op_type, node.inputs[0], node.outputs[0], err_message, self.coreml_3_rerun_message)
      )






from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import onnx
from typing import Any, Text, Optional, Tuple, Sequence
import onnx.backend.test
import numpy

from onnx_coreml._backend import CoreMLBackend, CoreMLBackendND

from coremltools.models.utils import macos_version

# Disable Rank 5 mapping for ONNX backend testing
DISABLE_RANK5_MAPPING = False

MIN_MACOS_VERSION_10_15 = (10, 15)
# If MACOS version is less than 10.15
# Then force testing on CoreML 2.0
if macos_version() < MIN_MACOS_VERSION_10_15:
    DISABLE_RANK5_MAPPING = False

class CoreMLTestingBackend(CoreMLBackendND if DISABLE_RANK5_MAPPING else CoreMLBackend):
    @classmethod
    def run_node(cls,
                 node,  # type: onnx.NodeProto
                 inputs,  # type: Sequence[numpy.ndarray[Any]]
                 device='CPU',  # type: Text
                 outputs_info=None,  # type: Optional[Sequence[Tuple[numpy.dtype, Tuple[int, ...]]]]
                 ):
        # type: (...) -> onnx.namedtupledict
        '''
        CoreML requires full model for prediction, not just single layer.
        Also input/output shapes are required to build CoreML spec for model.
        As a temporary decision we use caffe2 backend for shape inference
        task to build the appropriate ONNX model and convert it to
        CoreML model.
        '''
        super(CoreMLTestingBackend, cls).run_node(node, inputs, device)

        assert outputs_info is not None, "CoreML needs output shapes"

        graph_inputs = []
        for i in range(len(inputs)):
            input_ = inputs[i]
            value_info = onnx.helper.make_tensor_value_info(
                name=node.input[i],
                elem_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[input_.dtype],
                shape=input_.shape
            )
            graph_inputs.append(value_info)

        graph_outputs = []
        for i in range(len(outputs_info)):
            output_info = outputs_info[i]
            value_info = onnx.helper.make_tensor_value_info(
                name=node.output[i],
                elem_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[output_info[0]],
                shape=output_info[1]
            )
            graph_outputs.append(value_info)

        graph = onnx.helper.make_graph(
            nodes=[node],
            name='dummy',
            inputs=graph_inputs,
            outputs=graph_outputs,
        )

        model = onnx.helper.make_model(graph)
        return cls.prepare(model).run(inputs)


def exclude_test_cases(backend_test):
    # Test cases to be disabled till coreml 2
    unsupported_tests_till_coreml2 = [
        # Failing due to tolerance (in particular due to the way outputs are compared,
        'test_log_example_cpu',
        'test_operator_add_broadcast_cpu',
        'test_operator_add_size1_right_broadcast_cpu',
        'test_operator_add_size1_singleton_broadcast_cpu',
        'test_operator_addconstant_cpu',
        'test_sign_cpu',
        'test_sign_model_cpu',

        # Dynamic ONNX ops not supported in CoreML
        'test_reshape_extended_dims_cpu',
        'test_reshape_negative_dim_cpu',
        'test_reshape_one_dim_cpu',
        'test_reshape_reduced_dims_cpu',
        'test_reshape_reordered_dims_cpu',
        'test_gemm_broadcast_cpu',
        'test_gemm_nobroadcast_cpu',
        'test_upsample_nearest_cpu',

        # Failure due to axis alignment between ONNX and CoreML
        'test_add_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_div_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_mul_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_sub_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_operator_index_cpu', # input: [1,1]: cannot slice along batch axis
        'test_concat_2d_axis_0_cpu', # cannot slice along batch axis
        'test_argmax_default_axis_example_cpu', # batch axis
        'test_argmin_default_axis_example_cpu', # batch axis
        'test_AvgPool1d_cpu', # 3-D tensor unsqueezed to 4D with axis = 3 and then pool
        'test_AvgPool1d_stride_cpu', # same as above

        # Possibly Fixable by the converter
        'test_slice_start_out_of_bounds_cpu', # starts and ends exceed input size

        # "Supported ops, but Unsupported parameters by CoreML"
        'test_logsoftmax_axis_1_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_logsoftmax_default_axis_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_softmax_axis_1_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_softmax_default_axis_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_logsoftmax_axis_0_cpu',
        'test_logsoftmax_axis_2_cpu',
        'test_softmax_axis_0_cpu',
        'test_softmax_axis_2_cpu',
        'test_log_softmax_dim3_cpu',
        'test_log_softmax_lastdim_cpu',
        'test_softmax_functional_dim3_cpu',
        'test_PReLU_1d_multiparam_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_mvn_cpu', # not sure why there is numerical mismatch
        'test_flatten_axis2_cpu', # 2,3,4,5 --> 6,20
        'test_flatten_axis3_cpu', # 2,3,4,5 --> 24,5

        # Unsupported ops by CoreML
        'test_constant_cpu',
        'test_hardmax_axis_0_cpu',
        'test_hardmax_axis_1_cpu',
        'test_hardmax_axis_2_cpu',
        'test_hardmax_default_axis_cpu',
        'test_hardmax_example_cpu',
        'test_hardmax_one_hot_cpu',
        'test_matmul_2d_cpu',
        'test_matmul_3d_cpu',
        'test_matmul_4d_cpu',
        'test_AvgPool3d_cpu',
        'test_AvgPool3d_stride_cpu',
        'test_BatchNorm3d_eval_cpu',
        'test_Conv3d_cpu',
        'test_Conv3d_dilated_cpu',
        'test_Conv3d_dilated_strided_cpu',
        'test_Conv3d_groups_cpu',
        'test_Conv3d_no_bias_cpu',
        'test_Conv3d_stride_cpu',
        'test_Conv3d_stride_padding_cpu',
        'test_MaxPool3d_cpu',
        'test_MaxPool3d_stride_cpu',
        'test_MaxPool3d_stride_padding_cpu',
        'test_PReLU_3d_cpu',
        'test_PReLU_3d_multiparam_cpu',
        'test_AvgPool3d_stride1_pad0_gpu_input_cpu',
        'test_BatchNorm3d_momentum_eval_cpu',
        'test_BatchNorm1d_3d_input_eval_cpu',
        'test_averagepool_3d_default_cpu',
        'test_maxpool_3d_default_cpu',
        'test_split_variable_parts_1d_cpu',
        'test_split_variable_parts_2d_cpu',
        'test_split_variable_parts_default_axis_cpu',
        'test_cast_DOUBLE_to_FLOAT_cpu',
        'test_cast_FLOAT_to_DOUBLE_cpu',
        'test_shape_cpu',
        'test_shape_example_cpu',
        'test_size_cpu',
        'test_size_example_cpu',
        'test_top_k_cpu',
        'test_PoissonNLLLLoss_no_reduce_cpu',
        'test_operator_mm_cpu',
        'test_operator_addmm_cpu',
        'test_tile_cpu',
        'test_tile_precomputed_cpu',
        'test_acos_cpu',
        'test_acos_example_cpu',
        'test_asin_cpu',
        'test_asin_example_cpu',
        'test_atan_cpu',
        'test_atan_example_cpu',
        'test_cos_cpu',
        'test_cos_example_cpu',
        'test_sin_cpu',
        'test_sin_example_cpu',
        'test_tan_cpu',
        'test_tan_example_cpu',
        'test_dropout_cpu',
        'test_dropout_default_cpu',
        'test_dropout_random_cpu',
        'test_gru_seq_length_cpu',
        'test_identity_cpu',
        'test_asin_example_cpu',
        'test_reduce_log_sum_exp_default_axes_keepdims_example_cpu',
        'test_reduce_log_sum_exp_default_axes_keepdims_random_cpu',
        'test_reduce_log_sum_exp_do_not_keepdims_example_cpu',
        'test_reduce_log_sum_exp_do_not_keepdims_random_cpu',
        'test_reduce_log_sum_exp_keepdims_example_cpu',
        'test_reduce_log_sum_exp_keepdims_random_cpu',
        'test_rnn_seq_length_cpu',
        'test_operator_repeat_cpu',
        'test_operator_repeat_dim_overflow_cpu',
        'test_thresholdedrelu_example_cpu', #different convention for CoreML
        'test_expand_shape_model1_cpu',
        'test_expand_shape_model2_cpu',
        'test_expand_shape_model3_cpu',
        'test_expand_shape_model4_cpu',
        'test_expand_dim_changed_cpu',
        'test_expand_dim_unchanged_cpu',
        'test_operator_chunk_cpu', # unequal splits
        'test_operator_permute2_cpu', # rank 6 input
        'test_maxpool_with_argmax_2d_precomputed_pads_cpu',
        'test_maxpool_with_argmax_2d_precomputed_strides_cpu',
        'test_gather_1_cpu',
        'test_gather_0_cpu',
        'test_Embedding_cpu', # gather op
        'test_Embedding_sparse_cpu', # gather op
        'test_pow_bcast_scalar_cpu',
        'test_pow_bcast_array_cpu',
        'test_pow_cpu',
        'test_pow_example_cpu',
        'test_operator_pow_cpu',
        'test_shrink_cpu',
        'test_acosh_cpu',
        'test_asinh_cpu',
        'test_atanh_cpu',
        'test_erf_cpu',
        'test_isnan_cpu',
        'test_where_example_cpu',
        'test_acosh_example_cpu',
        'test_asinh_example_cpu',
        'test_atanh_example_cpu',
        'test_compress_0_cpu',
        'test_compress_1_cpu',
        'test_compress_default_axis_cpu',
        'test_constantofshape_float_ones_cpu',
        'test_constantofshape_int_zeros_cpu',
        'test_cosh_cpu',
        'test_cosh_example_cpu',
        'test_dynamic_slice_cpu',
        'test_dynamic_slice_default_axes_cpu',
        'test_dynamic_slice_end_out_of_bounds_cpu',
        'test_dynamic_slice_neg_cpu',
        'test_dynamic_slice_neg_cpu',
        'test_dynamic_slice_start_out_of_bounds_cpu',
        'test_eyelike_populate_off_main_diagonal_cpu',
        'test_eyelike_with_dtype_cpu',
        'test_eyelike_without_dtype_cpu',
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_scatter_with_axis_cpu',
        'test_scatter_without_axis_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',
        'test_basic_convinteger_cpu',


        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet',
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',
        'test_basic_convinteger_cpu',
        'test_cast_FLOAT_to_STRING_cpu',
        'test_cast_STRING_to_FLOAT_cpu',
        'test_cast_DOUBLE_to_FLOAT16_cpu',
        'test_cast_FLOAT16_to_DOUBLE_cpu',
        'test_cast_FLOAT16_to_FLOAT_cpu',
        'test_cast_FLOAT_to_FLOAT16_cpu',
        'test_matmulinteger_cpu',
        'test_prelu_broadcast_cpu',
        'test_convinteger_with_padding_cpu',
        'test_convtranspose_3d_cpu',
        'test_dequantizelinear_cpu',
        'test_instancenorm_epsilon_cpu',
        'test_instancenorm_example_cpu',
        'test_isinf_cpu',
        'test_isinf_negative_cpu',
        'test_isinf_positive_cpu',
        'test_nonmaxsuppression_center_point_box_format_cpu',
        'test_nonmaxsuppression_flipped_coordinates_cpu',
        'test_nonmaxsuppression_identical_boxes_cpu',
        'test_nonmaxsuppression_limit_output_size_cpu',
        'test_nonmaxsuppression_single_box_cpu',
        'test_nonmaxsuppression_suppress_by_IOU_and_scores_cpu',
        'test_nonmaxsuppression_suppress_by_IOU_cpu',
        'test_nonmaxsuppression_two_batches_cpu',
        'test_nonmaxsuppression_two_classes_cpu',
        'test_prelu_example_cpu',
        'test_qlinearconv_cpu',
        'test_qlinearmatmul_2D_cpu',
        'test_qlinearmatmul_3D_cpu',
        'test_quantizelinear_cpu',
        'test_resize_downsample_linear_cpu',
        'test_resize_downsample_nearest_cpu',
        'test_resize_nearest_cpu',
        'test_resize_upsample_linear_cpu',
        'test_resize_upsample_nearest_cpu',
        'test_reversesequence_batch_cpu',
        'test_reversesequence_time_cpu',
        'test_roialign_cpu',
        'test_slice_cpu',
        'test_slice_default_axes_cpu',
        'test_slice_default_steps_cpu',
        'test_slice_end_out_of_bounds_cpu',
        'test_slice_neg_cpu',
        'test_slice_neg_steps_cpu',
        'test_strnormalizer_export_monday_casesensintive_lower_cpu',
        'test_strnormalizer_export_monday_casesensintive_nochangecase_cpu',
        'test_strnormalizer_export_monday_casesensintive_upper_cpu',
        'test_strnormalizer_export_monday_empty_output_cpu',
        'test_strnormalizer_export_monday_insensintive_upper_twodim_cpu',
        'test_strnormalizer_nostopwords_nochangecase_cpu',
        'test_PReLU_1d_cpu',
        'test_PReLU_2d_cpu',
        'test_strnorm_model_monday_casesensintive_lower_cpu',
        'test_strnorm_model_monday_casesensintive_nochangecase_cpu',
        'test_strnorm_model_monday_casesensintive_upper_cpu',
        'test_strnorm_model_monday_empty_output_cpu',
        'test_strnorm_model_monday_insensintive_upper_twodim_cpu',
        'test_strnorm_model_nostopwords_nochangecase_cpu',
        'test_averagepool_2d_ceil_cpu',
        'test_convtranspose_pad_cpu',
        'test_logsoftmax_large_number_cpu',
        'test_maxpool_2d_ceil_cpu',
        'test_maxpool_2d_dilations_cpu',
        'test_softmax_large_number_cpu',
        'test_mod_bcast_cpu',
        'test_mod_fmod_mixed_sign_example_cpu',
        'test_mvn_expanded_cpu',
        'test_operator_non_float_params_cpu',
        'test_operator_params_cpu',

        # Incorrect test setup
        'test_mod_float_mixed_sign_example_cpu',
        'test_mod_int64_mixed_sign_example_cpu',
        
        # Need dead code elimination and removing unused inputs
        'test_batchnorm_epsilon_cpu',
        'test_batchnorm_example_cpu',
        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
    ]

    ## Test cases to be disabled in CoreML 3.0
    unsupported_tests_coreml3 = [
        # Failing due to tolerance (in particular due to the way outputs are compared,
        'test_log_example_cpu',
        'test_operator_add_broadcast_cpu',
        'test_operator_add_size1_right_broadcast_cpu',
        'test_operator_add_size1_singleton_broadcast_cpu',
        'test_operator_addconstant_cpu',
        'test_sign_cpu',
        'test_sign_model_cpu',

        # Dynamic ONNX ops not supported in CoreML
        'test_reshape_extended_dims_cpu',
        'test_reshape_negative_dim_cpu',
        'test_reshape_one_dim_cpu',
        'test_reshape_reduced_dims_cpu',
        'test_reshape_reordered_dims_cpu',
        'test_gemm_broadcast_cpu',
        'test_gemm_nobroadcast_cpu',
        'test_upsample_nearest_cpu',

        # Failure due to axis alignment between ONNX and CoreML
        'test_add_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_div_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_mul_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_sub_bcast_cpu', # (3,4,5, shaped tensor + (5,, shaped tensor
        'test_operator_index_cpu', # input: [1,1]: cannot slice along batch axis
        'test_concat_2d_axis_0_cpu', # cannot slice along batch axis
        'test_argmax_default_axis_example_cpu', # batch axis
        'test_argmin_default_axis_example_cpu', # batch axis
        'test_AvgPool1d_cpu', # 3-D tensor unsqueezed to 4D with axis = 3 and then pool
        'test_AvgPool1d_stride_cpu', # same as above

        # Possibly Fixable by the converter
        'test_slice_start_out_of_bounds_cpu', # starts and ends exceed input size

        # "Supported ops, but Unsupported parameters by CoreML"
        'test_logsoftmax_axis_1_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_logsoftmax_default_axis_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_softmax_axis_1_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_softmax_default_axis_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_logsoftmax_axis_0_cpu',
        'test_softmax_axis_0_cpu',
        'test_PReLU_1d_multiparam_cpu', # this one converts but fails at runtime: must give error during conversion
        'test_mvn_cpu', # not sure why there is numerical mismatch
        'test_flatten_axis2_cpu', # 2,3,4,5 --> 6,20
        'test_flatten_axis3_cpu', # 2,3,4,5 --> 24,5

        # Unsupported ops by CoreML
        'test_constant_cpu',
        'test_hardmax_axis_0_cpu',
        'test_hardmax_axis_1_cpu',
        'test_hardmax_axis_2_cpu',
        'test_hardmax_default_axis_cpu',
        'test_hardmax_example_cpu',
        'test_hardmax_one_hot_cpu',
        'test_matmul_2d_cpu',
        'test_matmul_3d_cpu',
        'test_matmul_4d_cpu',
        'test_AvgPool3d_cpu',
        'test_AvgPool3d_stride_cpu',
        'test_BatchNorm3d_eval_cpu',
        'test_Conv3d_cpu',
        'test_Conv3d_dilated_cpu',
        'test_Conv3d_dilated_strided_cpu',
        'test_Conv3d_groups_cpu',
        'test_Conv3d_no_bias_cpu',
        'test_Conv3d_stride_cpu',
        'test_Conv3d_stride_padding_cpu',
        'test_MaxPool3d_cpu',
        'test_MaxPool3d_stride_cpu',
        'test_MaxPool3d_stride_padding_cpu',
        'test_PReLU_3d_cpu',
        'test_PReLU_3d_multiparam_cpu',
        'test_AvgPool3d_stride1_pad0_gpu_input_cpu',
        'test_BatchNorm3d_momentum_eval_cpu',
        'test_BatchNorm1d_3d_input_eval_cpu',
        'test_averagepool_3d_default_cpu',
        'test_maxpool_3d_default_cpu',
        'test_split_variable_parts_1d_cpu',
        'test_split_variable_parts_2d_cpu',
        'test_split_variable_parts_default_axis_cpu',
        'test_shape_cpu',
        'test_shape_example_cpu',
        'test_top_k_cpu',
        'test_PoissonNLLLLoss_no_reduce_cpu',
        'test_operator_mm_cpu',
        'test_operator_addmm_cpu',
        'test_tile_cpu',
        'test_tile_precomputed_cpu',
        'test_acos_cpu',
        'test_acos_example_cpu',
        'test_asin_cpu',
        'test_asin_example_cpu',
        'test_atan_cpu',
        'test_atan_example_cpu',
        'test_cos_cpu',
        'test_cos_example_cpu',
        'test_sin_cpu',
        'test_sin_example_cpu',
        'test_tan_cpu',
        'test_tan_example_cpu',
        'test_dropout_cpu',
        'test_dropout_default_cpu',
        'test_dropout_random_cpu',
        'test_gru_seq_length_cpu',
        'test_identity_cpu',
        'test_asin_example_cpu',
        'test_reduce_log_sum_exp_default_axes_keepdims_example_cpu',
        'test_reduce_log_sum_exp_default_axes_keepdims_random_cpu',
        'test_reduce_log_sum_exp_do_not_keepdims_example_cpu',
        'test_reduce_log_sum_exp_do_not_keepdims_random_cpu',
        'test_reduce_log_sum_exp_keepdims_example_cpu',
        'test_reduce_log_sum_exp_keepdims_random_cpu',
        'test_rnn_seq_length_cpu',
        'test_operator_repeat_cpu',
        'test_operator_repeat_dim_overflow_cpu',
        'test_thresholdedrelu_example_cpu', #different convention for CoreML
        'test_expand_shape_model1_cpu',
        'test_expand_shape_model2_cpu',
        'test_expand_shape_model3_cpu',
        'test_expand_shape_model4_cpu',
        'test_expand_dim_changed_cpu',
        'test_operator_chunk_cpu', # unequal splits
        'test_operator_permute2_cpu', # rank 6 input
        'test_maxpool_with_argmax_2d_precomputed_pads_cpu',
        'test_maxpool_with_argmax_2d_precomputed_strides_cpu',
        'test_gather_1_cpu',
        'test_gather_0_cpu',
        'test_Embedding_cpu', # gather op
        'test_Embedding_sparse_cpu', # gather op
        'test_pow_bcast_scalar_cpu',
        'test_pow_bcast_array_cpu',
        'test_pow_cpu',
        'test_pow_example_cpu',
        'test_operator_pow_cpu',
        'test_shrink_cpu',
        'test_acosh_cpu',
        'test_asinh_cpu',
        'test_atanh_cpu',
        'test_erf_cpu',
        'test_isnan_cpu',
        'test_where_example_cpu',
        'test_acosh_example_cpu',
        'test_asinh_example_cpu',
        'test_atanh_example_cpu',
        'test_compress_0_cpu',
        'test_compress_1_cpu',
        'test_compress_default_axis_cpu',
        'test_constantofshape_float_ones_cpu',
        'test_constantofshape_int_zeros_cpu',
        'test_cosh_cpu',
        'test_cosh_example_cpu',
        'test_dynamic_slice_cpu',
        'test_dynamic_slice_default_axes_cpu',
        'test_dynamic_slice_end_out_of_bounds_cpu',
        'test_dynamic_slice_neg_cpu',
        'test_dynamic_slice_neg_cpu',
        'test_dynamic_slice_start_out_of_bounds_cpu',
        'test_eyelike_populate_off_main_diagonal_cpu',
        'test_eyelike_with_dtype_cpu',
        'test_eyelike_without_dtype_cpu',
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',
        'test_basic_convinteger_cpu',
        'test_cast_FLOAT_to_STRING_cpu',
        'test_cast_STRING_to_FLOAT_cpu',
        'test_cast_DOUBLE_to_FLOAT16_cpu',
        'test_cast_FLOAT16_to_DOUBLE_cpu',
        'test_cast_FLOAT16_to_FLOAT_cpu',
        'test_cast_FLOAT_to_FLOAT16_cpu',
        'test_matmulinteger_cpu',
        'test_prelu_broadcast_cpu',
        'test_convinteger_with_padding_cpu',
        'test_convtranspose_3d_cpu',
        'test_dequantizelinear_cpu',
        'test_instancenorm_epsilon_cpu',
        'test_instancenorm_example_cpu',
        'test_isinf_cpu',
        'test_isinf_negative_cpu',
        'test_isinf_positive_cpu',
        'test_nonmaxsuppression_center_point_box_format_cpu',
        'test_nonmaxsuppression_flipped_coordinates_cpu',
        'test_nonmaxsuppression_identical_boxes_cpu',
        'test_nonmaxsuppression_limit_output_size_cpu',
        'test_nonmaxsuppression_single_box_cpu',
        'test_nonmaxsuppression_suppress_by_IOU_and_scores_cpu',
        'test_nonmaxsuppression_suppress_by_IOU_cpu',
        'test_nonmaxsuppression_two_batches_cpu',
        'test_nonmaxsuppression_two_classes_cpu',
        'test_prelu_example_cpu',
        'test_qlinearconv_cpu',
        'test_qlinearmatmul_2D_cpu',
        'test_qlinearmatmul_3D_cpu',
        'test_quantizelinear_cpu',
        # Resize test fails due to unknown Scaling factor
        # Requires custom layer/function to bypass
        'test_resize_downsample_linear_cpu',
        'test_resize_downsample_nearest_cpu',
        'test_resize_nearest_cpu',
        'test_resize_upsample_linear_cpu',
        'test_resize_upsample_nearest_cpu',
        'test_reversesequence_time_cpu',
        'test_roialign_cpu',
        'test_slice_cpu',
        'test_slice_default_axes_cpu',
        'test_slice_default_steps_cpu',
        'test_slice_end_out_of_bounds_cpu',
        'test_slice_neg_cpu',
        'test_slice_neg_steps_cpu',
        'test_strnormalizer_export_monday_casesensintive_lower_cpu',
        'test_strnormalizer_export_monday_casesensintive_nochangecase_cpu',
        'test_strnormalizer_export_monday_casesensintive_upper_cpu',
        'test_strnormalizer_export_monday_empty_output_cpu',
        'test_strnormalizer_export_monday_insensintive_upper_twodim_cpu',
        'test_strnormalizer_nostopwords_nochangecase_cpu',
        'test_PReLU_1d_cpu',
        'test_PReLU_2d_cpu',
        'test_strnorm_model_monday_casesensintive_lower_cpu',
        'test_strnorm_model_monday_casesensintive_nochangecase_cpu',
        'test_strnorm_model_monday_casesensintive_upper_cpu',
        'test_strnorm_model_monday_empty_output_cpu',
        'test_strnorm_model_monday_insensintive_upper_twodim_cpu',
        'test_strnorm_model_nostopwords_nochangecase_cpu',
        'test_averagepool_2d_ceil_cpu',
        'test_convtranspose_pad_cpu',
        'test_logsoftmax_large_number_cpu',
        'test_maxpool_2d_ceil_cpu',
        'test_maxpool_2d_dilations_cpu',
        'test_softmax_large_number_cpu',

        # Incorrect test setup
        'test_mod_float_mixed_sign_example_cpu',
        'test_mod_int64_mixed_sign_example_cpu',
        
        # Need dead code elimination and removing unused inputs
        'test_batchnorm_epsilon_cpu',
        'test_batchnorm_example_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet',
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
        'test_maxunpool_export_with_output_shape_cpu',
        'test_maxunpool_export_without_output_shape_cpu',
        'test_nonzero_example_cpu',
        'test_onehot_with_axis_cpu',
        'test_onehot_without_axis_cpu',
        'test_scan9_sum_cpu',
        'test_scan_sum_cpu',
        'test_shrink_hard_cpu',
        'test_shrink_soft_cpu',
        'test_sinh_cpu',
        'test_sinh_example_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_batch_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_batch_uniandbigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_only_bigrams_skip0_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_levelempty_cpu',
        'test_tfidfvectorizer_tf_onlybigrams_skip5_cpu',
        'test_tfidfvectorizer_tf_uniandbigrams_skip5_cpu',

        # recurrent tests.
        'test_operator_rnn_cpu',
        'test_operator_rnn_single_layer_cpu',
        'test_operator_lstm_cpu',
        'test_gru_defaults_cpu',
        'test_gru_with_initial_bias_cpu',
        'test_lstm_defaults_cpu',
        'test_lstm_with_initial_bias_cpu',
        'test_lstm_with_peepholes_cpu',
        'test_simple_rnn_defaults_cpu',
        'test_simple_rnn_with_initial_bias_cpu',

        # exclude all the model zoo tests. They are tested elsewhere.
        'test_bvlc_alexnet',
        'test_resnet50',
        'test_vgg16',
        'test_vgg19',
        'test_densenet121',
        'test_inception_v1',
        'test_inception_v2',
        'test_shufflenet',
        'test_squeezenet',
        'test_zfnet'
    ]

    if DISABLE_RANK5_MAPPING:
        for each in unsupported_tests_coreml3:
            backend_test.exclude(each)
    else:
        for each in unsupported_tests_till_coreml2:
            backend_test.exclude(each)
    
# import all test cases at global scope to make them visible to python.unittest
backend_test = onnx.backend.test.BackendTest(CoreMLTestingBackend, __name__)

# exclude unsupported test cases
exclude_test_cases(backend_test)

globals().update(backend_test
                 .enable_report()
                 .test_cases)


if __name__ == '__main__':
    unittest.main()




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

from onnx import helper, numpy_helper, TensorProto

from tests._test_utils import _onnx_create_single_node_model, \
    _onnx_create_model, _conv_pool_output_size, _random_array

from onnx_coreml._graph import Node, Graph


class NodeTest(unittest.TestCase):
    def test_create_node(self):  # type: () -> None
        model = _onnx_create_single_node_model(
            "Elu",
            [(1, 3, 224, 224)],
            [(1, 3, 224, 224)],
            alpha=0.5
        )
        graph = model.graph
        node = graph.node[0]
        node_ = Node.from_onnx(node)
        self.assertTrue(len(node_.inputs) == 1)
        self.assertTrue(len(node_.outputs) == 1)
        self.assertTrue(len(node_.attrs) == 1)
        self.assertTrue(node_.attrs["alpha"] == 0.5)


class GraphTest(unittest.TestCase):
    def test_create_graph(self):  # type: () -> None
        kernel_shape = (3, 2)
        strides = (2, 3)
        pads = (4, 2, 4, 2)
        dilations = (1, 2)
        group = 1
        weight = numpy_helper.from_array(
            _random_array((16, 3, 3, 2)), name="weight"
        )

        input_shape = (1, 3, 224, 224)
        output_size = _conv_pool_output_size(input_shape, dilations,
                                             kernel_shape, pads, strides)

        output_shape = (1, int(weight.dims[0]), output_size[0], output_size[1])

        inputs = [('input0', input_shape)]
        outputs = [('output0', output_shape, TensorProto.FLOAT)]

        conv = helper.make_node(
            "Conv",
            inputs=[inputs[0][0], "weight"],
            outputs=["conv_output"],
            dilations=dilations,
            group=group,
            kernel_shape=kernel_shape,
            pads=pads,
            strides=strides
        )

        relu = helper.make_node(
            "Relu",
            inputs=[conv.output[0]],
            outputs=[outputs[0][0]]
        )

        model = _onnx_create_model([conv, relu], inputs, outputs, [weight])
        graph_ = Graph.from_onnx(model.graph)
        self.assertTrue(len(graph_.inputs) == 1)
        self.assertEqual(graph_.inputs[0][2], input_shape)
        self.assertTrue(len(graph_.outputs) == 1)
        self.assertEqual(graph_.outputs[0][2], output_shape)
        self.assertTrue(len(graph_.nodes) == 2)
        self.assertEqual(len(graph_.nodes[0].parents), 0)
        self.assertEqual(len(graph_.nodes[1].parents), 1)
        self.assertEqual(len(graph_.nodes[0].children), 1)
        self.assertEqual(len(graph_.nodes[1].children), 0)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from typing import Any, Text, Dict, Tuple
from onnx import ModelProto
from onnx.backend.base import Backend
from onnx_coreml._backend_rep import CoreMLRep
from onnx_coreml import convert
import onnx
from ._graph import _input_from_onnx_input, EdgeInfo

DEBUG = False

def _get_onnx_outputs_info(model): # type: (...) -> Dict[Text, EdgeInfo]
    """
    Takes in an onnx model and returns a dictionary 
    of onnx output names mapped to a tuple that is (output_name, type, shape)
    """
    if isinstance(model, str):
        onnx_model = onnx.load(model)
    elif isinstance(model, onnx.ModelProto):
        onnx_model = model

    graph = onnx_model.graph
    onnx_output_dict = {}
    for o in graph.output:
        out = _input_from_onnx_input(o)
        onnx_output_dict[out[0]] = out
    return onnx_output_dict


class CoreMLBackend(Backend):
    @classmethod
    def prepare(cls,
                model,  # type: ModelProto
                device='CPU',  # type: Text
                disable_rank5_mapping=False, # type: Bool
                **kwargs  # type: Any
                ):
        # type: (...) -> CoreMLRep
        super(CoreMLBackend, cls).prepare(model, device, **kwargs)
        if DEBUG:
            with open('/tmp/node_model.onnx', 'wb') as f:
                s = model.SerializeToString()
                f.write(s)
        coreml_model = convert(model, disable_coreml_rank5_mapping=disable_rank5_mapping)
        if DEBUG:
            coreml_model.save('/tmp/node_model.mlmodel')
        onnx_outputs_info = _get_onnx_outputs_info(model)
        return CoreMLRep(coreml_model, onnx_outputs_info, device == 'CPU', disable_rank5_mapping=disable_rank5_mapping)

    @classmethod
    def is_compatible(cls,
                       model,  # type: ModelProto
                       device='CPU',  # type: Text
                       **kwargs  # type: Any
                       ):  # type: (...) -> bool
         # Return whether the model is compatible with CoreML.
         '''
         This function will gradually grow to cover more cases. 
         Need to be careful of false negatives. There are some cases that seemingly 
         are not supported on CoreML, which the graph transformer optimizes and converts to 
         a graph that can be converted to CoreML. 
         
         1. Check whether the layers for which CoreML expects constant weights are in
            the list of initializers in the onnx graph
         2. unsupported ops like "And", "Or" etc
            
         '''

         node_set = set()
         initializer_set = set()
         graph = model.graph
         for t in graph.initializer:
             initializer_set.add(t.name)
         for node in graph.node:
             if node.op_type in ['ConvTranspose',
                                 'Conv',
                                 'BatchNormalization',
                                 'InstanceNormalization',
                                 'PRelu']:
                if len(node.input) > 1 and node.input[1] not in initializer_set:
                    return False
             node_set.add(node.op_type)

         # unsupported ops remove
         for node in graph.node:
             if node.op_type in ['Cast',
                                 'And',
                                 'Or',
                                 'Xor',
                                 'Not',
                                 'Less',
                                 'Greater',
                                 'Equal',
                                 'Ceil',
                                 'Floor']:
                 return False

         return True

    @classmethod
    def supports_device(cls,
                        device,  # type: Text
                        ):
        # type: (...) -> bool
        return device == 'CPU'


class CoreMLBackendND(Backend):
    @classmethod
    def prepare(cls,
                model,  # type: ModelProto
                device='CPU',  # type: Text
                disable_rank5_mapping=True, # type: Bool
                **kwargs  # type: Any
                ):
        # type: (...) -> CoreMLRep
        super(CoreMLBackendND, cls).prepare(model, device, **kwargs)
        if DEBUG:
            with open('/tmp/node_model.onnx', 'wb') as f:
                s = model.SerializeToString()
                f.write(s)
        coreml_model = convert(model, disable_coreml_rank5_mapping=disable_rank5_mapping)
        if DEBUG:
            coreml_model.save('/tmp/node_model.mlmodel')
        onnx_outputs_info = _get_onnx_outputs_info(model)
        return CoreMLRep(coreml_model, onnx_outputs_info, device == 'CPU', disable_rank5_mapping=disable_rank5_mapping)

    @classmethod
    def is_compatible(cls,
                       model,  # type: ModelProto
                       device='CPU',  # type: Text
                       **kwargs  # type: Any
                       ):  # type: (...) -> bool
        # Return whether the model is compatible with CoreML.
        '''
        This function will gradually grow to cover more cases. 
        Need to be careful of false negatives. There are some cases that seemingly 
        are not supported on CoreML, which the graph transformer optimizes and converts to 
        a graph that can be converted to CoreML. 
        
        2. Unsupported ops: If graph has one of unsupported op, exit
           
        '''
        ## TODO: Add un-supported ops
        unsupported_ops = []
        graph = model.graph
        for node in graph.node:
            if node.op_type in unsupported_ops:
                return False
        return True

    @classmethod
    def supports_device(cls,
                        device,  # type: Text
                        ):
        # type: (...) -> bool
        return device == 'CPU'



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import onnx

import onnx.backend.test

from onnx_coreml._backend import CoreMLBackend, CoreMLBackendND

from coremltools.models.utils import macos_version

# Disable Rank 5 mapping for ONNX backend testing
DISABLE_RANK5_MAPPING = False

MIN_MACOS_VERSION_10_15 = (10, 15)
# If MACOS version is less than 10.15
# Then force testing on CoreML 2.0
if macos_version() < MIN_MACOS_VERSION_10_15:
    DISABLE_RANK5_MAPPING = False

# import all test cases at global scope to make them visible to python.unittest
backend_test = onnx.backend.test.BackendTest(CoreMLBackendND if DISABLE_RANK5_MAPPING else CoreMLBackend, __name__)

# Only include the big models tests
backend_test.include('test_resnet50')
backend_test.include('test_inception_v1')
backend_test.include('test_inception_v2')
backend_test.include('test_densenet121')
backend_test.include('test_shufflenet')
backend_test.include('test_squeezenet')
backend_test.include('test_bvlc_alexnet')
backend_test.include('test_zfnet512')
backend_test.include('test_vgg19')

globals().update(backend_test
                 .enable_report()
                 .test_cases)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import numpy.testing as npt  # type: ignore
import numpy.random as npr
from onnx import helper, TensorProto, ModelProto, ValueInfoProto, TensorProto, NodeProto
from typing import Any, Sequence, Text, Tuple, Optional, Dict, List, TypeVar
from onnx_coreml import convert
from onnx_coreml._graph import Node
import sys
import shutil
import os

'''
0: dynamically generate random inputs, 
   use caffe2 backend for onnx and 
   also save out the generated input and output dicts for future use 
1: use the already saved out input and output dicts for testing
'''
TEST_MODE = 1

def _forward_onnx_model(model,  # type: ModelProto
                        input_dict,  # type: Dict[Text, np._ArrayLike[Any]]
                        test_name = '' # type: Text
                        ):
    # type: (...) -> np.ndarray[Any]
    if TEST_MODE:
        current_dir_path = os.path.dirname(os.path.realpath(__file__))
        loaded_obj = np.load(current_dir_path + '/test_data/' + test_name + '/output.npy', encoding='bytes', allow_pickle=True) #type: ignore
        out = loaded_obj.item()
    else:
        import caffe2.python.onnx.backend # type: ignore
        prepared_backend = caffe2.python.onnx.backend.prepare(model)
        out = prepared_backend.run(input_dict)
        out_dict = {}
        out_names = [v.name for v in model.graph.output]
        for out_name in out_names:
            out_dict[out_name] = out[out_name]
        dir = os.path.dirname(os.path.realpath(__file__)) + '/test_data/' + test_name + '/'
        np.save(dir + 'output.npy', out_dict) #type: ignore

    result = [out[v.name] for v in model.graph.output]
    output_shapes = [
        _shape_from_onnx_value_info(o) for o in model.graph.output
    ]
    for i, output in enumerate(result):
        result[i] = output.reshape(output_shapes[i])
    return np.array(result)


def _onnx_create_model(nodes,  # type: Sequence[NodeProto]
                       inputs,  # type: Sequence[Tuple[Text,Tuple[int, ...]]]
                       outputs,  # type: Sequence[Tuple[Text,Tuple[int, ...], int]]
                       initializer=[],  # type: Sequence[TensorProto]
                       ):
    # type: (...) -> ModelProto
    initializer_inputs = [
        helper.make_tensor_value_info(
            t.name,
            TensorProto.FLOAT,
            t.dims
        ) for t in initializer
    ]

    graph = helper.make_graph(
        nodes=nodes,
        name="test",
        inputs=initializer_inputs + [
            helper.make_tensor_value_info(
                input_[0],
                TensorProto.FLOAT,
                input_[1]
            ) for input_ in inputs
        ],
        outputs=[
            helper.make_tensor_value_info(
                output_[0],
                output_[2],
                output_[1]
            ) for output_ in outputs
        ],
        initializer=initializer
    )
    onnx_model = helper.make_model(graph)
    return onnx_model


def _onnx_create_single_node_model(op_type,  # type: Text
                                   input_shapes,  # type: Sequence[Tuple[int, ...]]
                                   output_shapes,  # type: Sequence[Tuple[int, ...]]
                                   initializer=[],  # type: Sequence[TensorProto]
                                   **kwargs  # type: Any
                                   ):
    # type: (...) -> ModelProto
    inputs = [
        ("input{}".format(i,), input_shapes[i])
        for i in range(len(input_shapes))
    ]
    outputs = [
        ("output{}".format(i,), output_shapes[i], TensorProto.FLOAT)
        for i in range(len(output_shapes))
    ]

    node = helper.make_node(
        op_type,
        inputs=[i[0] for i in inputs] + [t.name for t in initializer],
        outputs=[o[0] for o in outputs],
        **kwargs
    )
    return _onnx_create_model([node], inputs, outputs, initializer)


def _shape_from_onnx_value_info(v):  # type: (ValueInfoProto) -> Sequence[Tuple[int, ...]]
    return tuple([d.dim_value for d in v.type.tensor_type.shape.dim])

def _coreml_forward_model(model,  # type: ModelProto
                          input_dict,  # type: Dict[Text, np._ArrayLike[Any]]
                          output_names,  # type: Sequence[Text]
                          disable_rank5_mapping=False
                          ):
    # type: (...) -> np.ndarray[Any]
    if not disable_rank5_mapping:
        for k, arr in input_dict.items():
            if len(arr.shape) == 4:
                input_dict[k] = arr[0]
        for k,v in input_dict.items():
            if len(v.shape) == 2 and v.shape[0] == 1:
                input_dict[k] = v.flatten()
    coreml_out = model.predict(input_dict, useCPUOnly=True)
    return np.array([coreml_out[name] for name in output_names])


def _coreml_forward_onnx_model(model,  # type: ModelProto
                               input_dict,  # type: Dict[Text, np._ArrayLike[Any]]
                               onnx_coreml_input_shape_map = {}, # type: Dict[Text, List[int,...]]
                               disable_rank5_mapping=False
                               ):
    # type: (...) -> np.ndarray[Any]
    coreml_model = convert(model, onnx_coreml_input_shape_map=onnx_coreml_input_shape_map, disable_coreml_rank5_mapping=disable_rank5_mapping)
    output_names = [o.name for o in model.graph.output]
    return _coreml_forward_model(coreml_model, input_dict, output_names, disable_rank5_mapping=disable_rank5_mapping)


def _random_array(shape, random_seed=10):  # type: (Tuple[int, ...], Any) -> np._ArrayLike[float]
    if random_seed:
        npr.seed(random_seed) # type: ignore
    return npr.ranf(shape).astype("float32")


def _conv_pool_output_size(input_shape,  # type: Sequence[int]
                           dilations,  # type: Sequence[int]
                           kernel_shape,  # type: Tuple[int, int]
                           pads,  # type: Sequence[int]
                           strides,  # type: Tuple[int, int]
                           ):
    # type: (...) -> Tuple[int, int]
    output_height = (
        input_shape[2] + pads[0] + pads[2] -
        (dilations[0] * (kernel_shape[0] - 1) + 1)
    ) / strides[0] + 1
    output_width = (
        input_shape[3] + pads[1] + pads[3] -
        (dilations[1] * (kernel_shape[1] - 1) + 1)
    ) / strides[1] + 1

    return (int(output_height), int(output_width))


_T = TypeVar('_T')


def _assert_outputs(output1,  # type: np.ndarray[_T]
                    output2,  # type: np.ndarray[_T]
                    decimal=7,  # type: int
                    ):
    # type: (...) -> None
    npt.assert_equal(len(output1), len(output2))
    for o1, o2 in zip(output1, output2):
        npt.assert_almost_equal(
            o2.flatten(),
            o1.flatten(),
            decimal=decimal
        )


def _prepare_inputs_for_onnx(model,  # type: ModelProto
                             test_name = '', # type: Text
                             values=None,  # type: Optional[List[np._ArrayLike[Any]]]
                             ):
    # type: (...) -> Dict[Text, np._ArrayLike[Any]]
    graph = model.graph
    initializer_names = {t.name for t in graph.initializer}
    input_names = [
        i.name for i in graph.input if i.name not in initializer_names
    ]
    input_shapes = [
        tuple([d.dim_value for d in i.type.tensor_type.shape.dim])
        for i in graph.input if i.name not in initializer_names
    ]

    if TEST_MODE:
        dir_path = os.path.dirname(os.path.realpath(__file__))
        loaded_obj = np.load(dir_path + '/test_data/' + test_name + '/input.npy', encoding='bytes', allow_pickle=True) # type: ignore
        return loaded_obj.item() # type: ignore
    else:
        if values is None:
            inputs = [_random_array(shape) for shape in input_shapes]
        else:
            inputs = values
        input_dict = dict(zip(input_names, inputs))
        dir = os.path.dirname(os.path.realpath(__file__)) + '/test_data/' + test_name + '/'
        if os.path.exists(dir):
            shutil.rmtree(dir)
        os.makedirs(dir)
        np.save(dir + 'input.npy', input_dict) # type: ignore
        return input_dict


def _test_onnx_model(model,  # type: ModelProto
                     test_name='', # type: Text
                     decimal=5,  # type: int
                     onnx_coreml_input_shape_map = {}, # type: Dict[Text, List[int,...]]
                     coreml_input_shape = {}, # type: Dict[Text, List[int,...]]
                     disable_rank5_mapping=False
                     ):
    # type: (...) -> None
    if not test_name:
        test_name = sys._getframe(1).f_code.co_name
    W = _prepare_inputs_for_onnx(model, test_name=test_name)
    c2_outputs = _forward_onnx_model(model, W, test_name=test_name)
    coreml_input_dict = dict()
    for key, value in W.items():
        if not disable_rank5_mapping and key in coreml_input_shape:
            coreml_input_dict[key] = np.reshape(value, coreml_input_shape[key])
        else:
            coreml_input_dict[key] = value
    coreml_outputs = _coreml_forward_onnx_model(model, coreml_input_dict, onnx_coreml_input_shape_map=onnx_coreml_input_shape_map,
                                                disable_rank5_mapping=disable_rank5_mapping)
    _assert_outputs(c2_outputs, coreml_outputs, decimal=decimal)


def _test_single_node(op_type,  # type: Text
                      input_shapes,  # type: Sequence[Tuple[int, ...]]
                      output_shapes,  # type: Sequence[Tuple[int, ...]]
                      initializer=[],  # type: Sequence[TensorProto]
                      decimal=5,  # type: int
                      test_name = '', # type: Text
                      onnx_coreml_input_shape_map = {}, # type: Dict[Text, List[int,...]]
                      coreml_input_shape = {}, # type: Dict[Text, List[int,...]]
                      disable_rank5_mapping=False,
                      **kwargs  # type: Any
                      ):
    # type: (...) -> None
    model = _onnx_create_single_node_model(
        op_type, input_shapes, output_shapes, initializer, **kwargs
    )
    if not test_name:
        test_name = sys._getframe(1).f_code.co_name
    _test_onnx_model(model, test_name=test_name, decimal=decimal,
                     onnx_coreml_input_shape_map=onnx_coreml_input_shape_map,
                     coreml_input_shape = coreml_input_shape,
                     disable_rank5_mapping=disable_rank5_mapping)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from typing import Text, Union, Optional, Dict, Any, Iterable, Sequence, Callable, List

import onnx
import numpy as np

from onnx import shape_inference
from onnx import TensorProto

from coremltools.models.neural_network import NeuralNetworkBuilder  #type: ignore
from coremltools.models import datatypes, MLModel  #type: ignore
from coremltools.proto import FeatureTypes_pb2 as ft  #type: ignore

from typing import Tuple

from ._operators import _convert_node, _SEQUENCE_LAYERS_REGISTRY, _ONNX_NODE_REGISTRY, _add_const_inputs_if_required
from ._operators_nd import _ONNX_NODE_REGISTRY_ND, _convert_node_nd

from ._graph import Graph, EdgeInfo, Transformer

from ._transformers import ConvAddFuser, DropoutRemover, \
    ReshapeInitTensorFuser, BNBroadcastedMulFuser, BNBroadcastedAddFuser, \
    PixelShuffleFuser, OutputRenamer, AddModelInputsOutputs, \
    ConstantsToInitializers, ImageScalerRemover, UnsqueezeConstantRemover, TransposeConstantRemover, \
    ShapeOpRemover, SliceConstantRemover, ConcatConstantRemover, DivMulConstantRemover, GatherConstantRemover, \
    ConstantFillToInitializers, ReshapeTransposeReshape_pattern1

from ._error_utils import ErrorHandling
from .graph_viz import plot_graph # type: ignore

USE_SHAPE_MAPPING = True

DEBUG = False

'''
inputs: list of tuples.
      [Tuple]: [(name, type, shape)]
'''
def _make_coreml_input_features(graph, onnx_coreml_input_shape_map, disable_coreml_rank5_mapping=False): # type: (...) -> Sequence[Tuple[Text, datatypes.Array]]
    '''
    If "disable_coreml_rank5_mapping" is False, then:

    ONNX shapes to CoreML static shapes mapping
    length==1: [C]
    length==2: [B,C]
    length==3: [C,H,W] or [Seq,B,C]
    length==4: [B,C,H,W]

    If "disable_coreml_rank5_mapping" is True, then
    onnx shapes are mapped "as is" to CoreML.
    '''
    inputs = graph.inputs
    op_types = graph.blob_to_op_type
    features = []
    for input_ in inputs:
        shape = input_[2]
        if disable_coreml_rank5_mapping:
            if len(shape) > 5:
                raise ValueError('ONNX input %s has a rank greater than 5, which is not supported in CoreML framework' % str(input_[0]))
            else:
                features.append((str(input_[0]), datatypes.Array(*shape)))
            continue

        if USE_SHAPE_MAPPING and input_[0] in onnx_coreml_input_shape_map:
            mapp = onnx_coreml_input_shape_map[input_[0]]
            if len(mapp) != len(shape):
                raise ValueError('Incorrect value in onnx_coreml_input_shape_map argument')
            graph.onnx_coreml_shape_mapping[input_[0]] = mapp
            coreml_shape = [1,1,1]
            for i in range(3):
                if (i+2) in mapp:
                    coreml_shape[i] = shape[mapp.index(i+2)]
            shape = coreml_shape
        else:
            if len(shape) == 0:
                shape = [1, 1, 1]
            elif len(shape) == 1:
                # assume [C]
                if USE_SHAPE_MAPPING:
                    graph.onnx_coreml_shape_mapping[input_[0]] = [2]
            elif len(shape) == 2:
                # assume [Batch,C]
                shape = [shape[1]]
                if USE_SHAPE_MAPPING:
                    graph.onnx_coreml_shape_mapping[input_[0]] = [1,2]
            elif len(shape) == 3:
                # assume [C,H,W] unless its connected an op that bestows another mapping
                if input_[0] in op_types and len(op_types[input_[0]]) == 1:
                    if str(op_types[input_[0]][0]) in _SEQUENCE_LAYERS_REGISTRY:
                        # (Seq,B,C)
                        shape = [shape[2]]
                        if USE_SHAPE_MAPPING:
                            graph.onnx_coreml_shape_mapping[input_[0]] = [0, 1, 2]
                    elif str(op_types[input_[0]][0]) in ['MaxPool','AveragePool','BatchNormalization',
                                                         'GlobalAveragePool','GlobalLpPool','GlobalMaxPool',
                                                         'InstanceNormalization','LRN','LpPool','Conv','ConvTranspose']:
                        # (B,C,W)
                        shape = [shape[1],1,shape[2]]
                        if USE_SHAPE_MAPPING:
                            graph.onnx_coreml_shape_mapping[input_[0]] = [1, 2, 4]
                    else:
                        if USE_SHAPE_MAPPING:
                            graph.onnx_coreml_shape_mapping[input_[0]] = [2, 3, 4]
                else:
                    if USE_SHAPE_MAPPING:
                        graph.onnx_coreml_shape_mapping[input_[0]] = [2, 3, 4]
            elif len(shape) == 4:  # (B,C,H,W) --> (C,H,W)
                shape = shape[1:]
                if USE_SHAPE_MAPPING:
                    graph.onnx_coreml_shape_mapping[input_[0]] = [1,2,3,4]
            else:
                raise ValueError("CoreML input cannot be more than rank 4. Input shape: %s, input: '%s' " % (str(shape), str(input_[0])))
        features.append((str(input_[0]), datatypes.Array(*shape)))
    return features

'''
outputs: list of tuples.
      [Tuple]: [(name, type, shape)]
'''
def _make_coreml_output_features(graph, forceShape=False, disable_coreml_rank5_mapping=False):  # type: (...) -> Sequence[Tuple[Text, datatypes.Array]]
    features = []
    outputs = graph.outputs
    op_types = graph.blob_from_op_type
    ops_allowing_zerod_output = {'Size'}

    for output_ in outputs:
        if op_types[output_[0]] in ops_allowing_zerod_output and len(output_[2]) == 0:
            output_ = list(output_)
            output_[2] = (1,)

        if disable_coreml_rank5_mapping:
            shape = output_[2]
            if len(shape) > 5:
                raise ValueError('ONNX output %s has a rank greater than 5, which is not supported in CoreML framework' % str(output_[0]))
            else:
                features.append((str(output_[0]), datatypes.Array(*shape)))
            continue

        if not forceShape:
            features.append((str(output_[0]), None))
        else:
            shape = output_[2]
            if len(shape) == 0:
                shape = [1, 1, 1]
            elif len(shape) == 1:
                pass
            elif len(shape) == 3:
                if output_[0] in op_types and \
                        str(op_types[output_[0]]) in _SEQUENCE_LAYERS_REGISTRY:
                    # onnx shape: (Seq,B,C)
                    shape = [shape[2]]
            elif len(shape) == 4:  # (B,C,H,W) --> (C,H,W)
                shape = shape[1:]
            else:
                shape = None  # output shape need not be specified for CoreML.
            if shape is None:
                features.append((str(output_[0]), shape))
            else:
                features.append((str(output_[0]), datatypes.Array(*shape)))
    return features

def _check_unsupported_ops(nodes, disable_coreml_rank5_mapping=False): # type: (...) -> None
    unsupported_op_types = [] # type: List[Text]
    for node in nodes:

        if disable_coreml_rank5_mapping:
            if node.op_type not in _ONNX_NODE_REGISTRY_ND and \
                    node.op_type not in unsupported_op_types:
                unsupported_op_types.append(node.op_type)
            continue

        if node.op_type not in _ONNX_NODE_REGISTRY and \
          node.op_type not in unsupported_op_types:
            unsupported_op_types.append(node.op_type)

    coreml_3_rerun_message = ''
    if not disable_coreml_rank5_mapping:
        coreml_3_rerun_message = '\nPlease try converting again with disable_coreml_rank5_mapping=True' \
                                 ' and coremltools 3.0 latest beta'
    if len(unsupported_op_types) > 0:
        raise NotImplementedError("Unsupported ONNX ops of type: %s %s" % (
            ','.join(unsupported_op_types), coreml_3_rerun_message))


def _update_multiarray_to_float32(feature, #type: Any
                                 ): # type : (...) -> None
  if feature.type.HasField('multiArrayType'):
    feature.type.multiArrayType.dataType = ft.ArrayFeatureType.FLOAT32

def _update_multiarray_to_int32(feature, #type: Any
                               ): # type : (...) -> None
  if feature.type.HasField('multiArrayType'):
    feature.type.multiArrayType.dataType = ft.ArrayFeatureType.INT32


def _transform_coreml_dtypes(builder, # type : NeuralNetworkBuilder
                             inputs, # type: List[EdgeInfo]
                             outputs # type: List[EdgeInfo]
                             ):
    # type: (...) -> None

    ''' Make sure ONNX input/output data types are mapped to the equivalent CoreML types
    '''
    for i, input_ in enumerate(inputs):
        onnx_type = input_[1]
        if onnx_type == TensorProto.FLOAT:
            _update_multiarray_to_float32(builder.spec.description.input[i])
        elif onnx_type == TensorProto.DOUBLE:
            continue
        elif onnx_type == TensorProto.INT32 or onnx_type == TensorProto.INT64:
            _update_multiarray_to_int32(builder.spec.description.input[i])
        elif onnx_type == TensorProto.BOOL:
            _update_multiarray_to_float32(builder.spec.description.input[i])
        else:
            raise TypeError("Input must be of of type FLOAT, DOUBLE, INT32 or INT64")

    for i, output_ in enumerate(outputs):
        onnx_type = output_[1]
        if onnx_type == TensorProto.FLOAT:
            _update_multiarray_to_float32(builder.spec.description.output[i])
        elif onnx_type == TensorProto.DOUBLE:
            continue
        elif onnx_type == TensorProto.INT32 or onnx_type == TensorProto.INT64:
            _update_multiarray_to_int32(builder.spec.description.output[i])
        elif onnx_type == TensorProto.BOOL:
            _update_multiarray_to_float32(builder.spec.description.output[i])
        else:
            raise TypeError("Output must be of of type FLOAT, DOUBLE, INT32 or INT64")

def _convert_multiarray_output_to_image(spec,  # type: Any
                                        feature_name,  # type: Text
                                        is_bgr=False,  # type: bool
                                        ):
    # type: (...) -> None
    for output in spec.description.output:
        if output.name != feature_name:
            continue
        if output.type.WhichOneof('Type') != 'multiArrayType':
            raise ValueError(
                "{} is not a multiarray type".format(output.name,)
            )
        array_shape = tuple(output.type.multiArrayType.shape)
        if len(array_shape) == 2:
            height, width = array_shape
            output.type.imageType.colorSpace = \
                ft.ImageFeatureType.ColorSpace.Value('GRAYSCALE')
        else:
            if len(array_shape) == 4:
                if array_shape[0] != 1:
                    raise ValueError(
                        "Shape {} is not supported for image output"
                        .format(array_shape,)
                    )
                array_shape = array_shape[1:]

            channels, height, width = array_shape

            if channels == 1:
                output.type.imageType.colorSpace = \
                    ft.ImageFeatureType.ColorSpace.Value('GRAYSCALE')
            elif channels == 3:
                if is_bgr:
                    output.type.imageType.colorSpace = \
                        ft.ImageFeatureType.ColorSpace.Value('BGR')
                else:
                    output.type.imageType.colorSpace = \
                        ft.ImageFeatureType.ColorSpace.Value('RGB')
            else:
                raise ValueError(
                    "Channel Value {} is not supported for image output"
                    .format(channels,)
                )

        output.type.imageType.width = width
        output.type.imageType.height = height


def _set_deprocessing(is_grayscale,  # type: bool
                      builder,  # type: NeuralNetworkBuilder
                      deprocessing_args,  # type: Dict[Text, Any]
                      input_name,  # type: Text
                      output_name,  # type: Text
                      ):
    # type: (...) -> None
    is_bgr = deprocessing_args.get('is_bgr', False)

    image_scale = deprocessing_args.get('image_scale', 1.0)

    if is_grayscale:
        gray_bias = deprocessing_args.get('gray_bias', 0.0)
        W = np.array([image_scale])
        b = np.array([gray_bias])
    else:
        W = np.array([image_scale, image_scale, image_scale])

        red_bias = deprocessing_args.get('red_bias', 0.0)
        green_bias = deprocessing_args.get('green_bias', 0.0)
        blue_bias = deprocessing_args.get('blue_bias', 0.0)

        if not is_bgr:
            b = np.array([
                red_bias,
                green_bias,
                blue_bias,
            ])
        else:
            b = np.array([
                blue_bias,
                green_bias,
                red_bias,
            ])
    builder.add_scale(
        name=input_name,
        W=W,
        b=b,
        has_bias=True,
        shape_scale=W.shape,
        shape_bias=b.shape,
        input_name=input_name,
        output_name=output_name
    )


def _prepare_onnx_graph(graph, transformers):  # type: (Graph, Iterable[Transformer]) -> Graph
    graph_ = Graph.from_onnx(graph)
    if DEBUG:
        plot_graph(graph_, graph_img_path='/tmp/graph_raw.pdf')
    graph_ = graph_.transformed(transformers)
    if DEBUG:
        plot_graph(graph_, graph_img_path='/tmp/graph_opt.pdf')
    return graph_

def convert(model,  # type: Union[onnx.ModelProto, Text]
            mode=None,  # type: Optional[Text]
            image_input_names=[],  # type: Sequence[Text]
            preprocessing_args={},  # type: Dict[Text, Any]
            image_output_names=[],  # type: Sequence[Text]
            deprocessing_args={},  # type: Dict[Text, Any]
            class_labels=None,  # type: Union[Text, Iterable[Text], None]
            predicted_feature_name='classLabel',  # type: Text
            add_custom_layers = False,  # type: bool
            custom_conversion_functions = {}, #type: Dict[Text, Any]
            onnx_coreml_input_shape_map = {}, # type: Dict[Text, List[int,...]]
            disable_coreml_rank5_mapping = False):
    # type: (...) -> MLModel
    """
    Convert ONNX model to CoreML.
    Parameters
    ----------
    model:
        An ONNX model with parameters loaded in onnx package or path to file
        with models.
    mode: 'classifier', 'regressor' or None
        Mode of the converted coreml model:
        'classifier', a NeuralNetworkClassifier spec will be constructed.
        'regressor', a NeuralNetworkRegressor spec will be constructed.
    preprocessing_args:
        'is_bgr', 'red_bias', 'green_bias', 'blue_bias', 'gray_bias',
        'image_scale' keys with the same meaning as
        https://apple.github.io/coremltools/generated/coremltools.models.neural_network.html#coremltools.models.neural_network.NeuralNetworkBuilder.set_pre_processing_parameters
    deprocessing_args:
        Same as 'preprocessing_args' but for deprocessing.
    class_labels:
        As a string it represents the name of the file which contains
        the classification labels (one per line).
        As a list of strings it represents a list of categories that map
        the index of the output of a neural network to labels in a classifier.
    predicted_feature_name:
        Name of the output feature for the class labels exposed in the Core ML
        model (applies to classifiers only). Defaults to 'classLabel'
    add_custom_layers: bool
        Flag to turn on addition of custom CoreML layers for unsupported ONNX ops or attributes within
        a supported op.
    custom_conversion_functions: dict()
        A dictionary with keys corresponding to the names/types of onnx ops and values as functions taking
        an object of class coreml-tools's 'NeuralNetworkBuilder', Graph' (see onnx-coreml/_graph.Graph),
        'Node' (see onnx-coreml/_graph.Node), ErrorHandling (see onnx-coreml/_error_utils.ErrorHandling).
        This custom conversion function gets full control and responsibility for converting given onnx op.
        This function returns nothing and is responsible for adding a equivalent CoreML layer via 'NeuralNetworkBuilder'
    onnx_coreml_input_shape_map: dict()
        (Optional) A dictionary with keys corresponding to the model input names. Values are a list of integers that specify
        how the shape of the input is mapped to CoreML. Convention used for CoreML shapes is
        0: Sequence, 1: Batch, 2: channel, 3: height, 4: width.
        For example, an input of rank 2 could be mapped as [3,4] (i.e. H,W) or [1,2] (i.e. B,C) etc.
        This is ignored if "disable_coreml_rank5_mapping" is set to True.
    disable_coreml_rank5_mapping: bool
        If True, then it disables the "RANK5_ARRAY_MAPPING" or enables the "EXACT_ARRAY_MAPPING"
        option in CoreML (https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L67)
        Thus, no longer, onnx tensors are forced to map to rank 5 CoreML tensors.
        With this flag on, a rank r ONNX tensor, (1<=r<=5), will map to a rank r tensor in CoreML as well.
        This flag must be on to utilize any of the new layers added in CoreML 3 (i.e. specification version 4, iOS13)

    Returns
    -------
    model: A coreml model.
    """
    if isinstance(model, Text):
        onnx_model = onnx.load(model)
    elif isinstance(model, onnx.ModelProto):
        onnx_model = model
    else:
        raise TypeError(
            "Model must be file path to .onnx file or onnx loaded model"
        )

    global USE_SHAPE_MAPPING
    if disable_coreml_rank5_mapping:
        USE_SHAPE_MAPPING = False
    else:
        USE_SHAPE_MAPPING = True


    '''
    First, apply a few optimizations to the ONNX graph,
    in preparation for conversion to CoreML. 
    '''

    # Using Dummy transformation to conditionally disable certain transformation
    class  DummyTransformation(object):
        def __call__(self, graph):
            return graph

    transformers = [
        ConstantsToInitializers(),
        ShapeOpRemover(),
        ReshapeInitTensorFuser(),
        DropoutRemover(),
        UnsqueezeConstantRemover(),
        TransposeConstantRemover(),
        SliceConstantRemover(),
        ConcatConstantRemover(),
        ConvAddFuser(),
        BNBroadcastedMulFuser(),
        BNBroadcastedAddFuser(),
        ReshapeTransposeReshape_pattern1(),
        PixelShuffleFuser(),
        AddModelInputsOutputs() if not disable_coreml_rank5_mapping else DummyTransformation(),
        DivMulConstantRemover(),
        GatherConstantRemover(),
        ConstantFillToInitializers(),
    ]  # type: Iterable[Transformer]


    onnx_model = onnx.shape_inference.infer_shapes(onnx_model)
    graph = _prepare_onnx_graph(onnx_model.graph, transformers)

    '''
    Check for ImageScalar nodes in ONNX, this will indicate whether input image preprocessing needs
    to be added to the CoreML graph or not. 
    '''
    # are there ImageScaler nodes in the Graph?
    # If yes then add the info from it to the "preprocessing_args" dictionary, if the dictionary is not
    # already provided by the user
    if not bool(preprocessing_args):
        for node in graph.nodes:
            if node.op_type == 'ImageScaler':
                inp_name = node.inputs[0]
                scale = node.attrs.get('scale', 1.0)
                bias = node.attrs.get('bias', [0,0,0])
                if not (len(bias) == 1 or len(bias) == 3):
                    continue
                if 'image_scale' in preprocessing_args:
                    preprocessing_args['image_scale'][inp_name] = scale
                else:
                    preprocessing_args['image_scale'] = {inp_name: scale}
                if len(bias) == 3:
                    for i, color in enumerate(['red', 'green', 'blue']):
                        if color + '_bias' in preprocessing_args:
                            preprocessing_args[color + '_bias'][inp_name] = bias[i]
                        else:
                            preprocessing_args[color + '_bias'] = {inp_name: bias[i]}
                else:
                    if 'gray_bias' in preprocessing_args:
                        preprocessing_args['gray_bias'][inp_name] = bias[0]
                    else:
                        preprocessing_args['gray_bias'] = {inp_name: bias[0]}
                if inp_name not in image_input_names:
                    image_input_names.append(inp_name) # type: ignore

    # remove all ImageScaler ops
    graph = graph.transformed([ImageScalerRemover()])

    '''
    Gather information (name, shape) for model inputs and outputs
    This information is then used to initialize the neural network builder object of coremltools. 
    The builder object is later used to add layers to the CoreML model. 
    '''

    #Make CoreML input and output features by gathering shape info and
    #interpreting it for CoreML
    input_features = _make_coreml_input_features(graph, onnx_coreml_input_shape_map, disable_coreml_rank5_mapping)
    if len( image_output_names) > 0:
        output_features = _make_coreml_output_features(graph, forceShape=True, disable_coreml_rank5_mapping=disable_coreml_rank5_mapping)
    else:
        output_features = _make_coreml_output_features(graph, disable_coreml_rank5_mapping=disable_coreml_rank5_mapping)

    builder = NeuralNetworkBuilder(input_features, output_features, mode=mode, disable_rank5_shape_mapping=disable_coreml_rank5_mapping)

    '''
    Set CoreML input,output types (float, double, int) same as onnx types, if supported
    '''
    _transform_coreml_dtypes(builder, graph.inputs, graph.outputs)


    '''what follows is some book-keeping to support outputs of type image. 
    '''

    is_deprocess_bgr_only = (len(deprocessing_args) == 1) and \
                            ("is_bgr" in deprocessing_args)
    add_deprocess = (len(image_output_names) > 0) and \
                    (len(deprocessing_args) > 0) and \
                    (not is_deprocess_bgr_only)

    if add_deprocess:
        mapping = {}
        for f in output_features:
            output_name = f[0]
            mapping[output_name] = graph.get_unique_edge_name(output_name)
        graph = OutputRenamer(mapping)(graph)



    if len(image_input_names) > 0:
        builder.set_pre_processing_parameters(
            image_input_names=image_input_names,
            is_bgr=preprocessing_args.get('is_bgr', False),
            red_bias=preprocessing_args.get('red_bias', 0.0),
            green_bias=preprocessing_args.get('green_bias', 0.0),
            blue_bias=preprocessing_args.get('blue_bias', 0.0),
            gray_bias=preprocessing_args.get('gray_bias', 0.0),
            image_scale=preprocessing_args.get('image_scale', 1.0)
        )

    preprocessing_args.clear()

    if len(image_output_names) > 0:
        for f in output_features:
            f_name = f[0]
            if f_name in image_output_names:
                is_bgr = deprocessing_args.get('is_bgr', False)
                _convert_multiarray_output_to_image(
                    builder.spec, f_name, is_bgr=is_bgr
                )

    '''
    Iterate through all the ONNX ops and translate them to CoreML layers, one by one. 
    '''

    '''
    before proceeding to start the layer translation process,
    check whether there is an op in the ONNX graph, whose translation function is not yet
    implemented in the converter or which is not supported in the CoreML framework. If so, 
    raise an error before starting the process.
    (if the user desires to add a custom layer then this check is not required)
    '''
    if not add_custom_layers:
        _check_unsupported_ops(graph.nodes, disable_coreml_rank5_mapping)

    '''
    ErrorHandling is a generic class, useful to store a variety of parameters during the conversion process  
    '''
    err = ErrorHandling(add_custom_layers,
                        custom_conversion_functions,
                        disable_coreml_rank5_mapping=disable_coreml_rank5_mapping)


    for i, node in enumerate(graph.nodes):
        print("%d/%d: Converting Node Type %s" %(i+1, len(graph.nodes), node.op_type))
        if disable_coreml_rank5_mapping:
            _convert_node_nd(builder, node, graph, err)
        else:
            _add_const_inputs_if_required(builder, node, graph, err)
            _convert_node(builder, node, graph, err)

    if DEBUG:
        plot_graph(graph, graph_img_path='/tmp/after_conversion.pdf', show_coreml_mapped_shapes=not disable_coreml_rank5_mapping) 

    if add_deprocess:
        for f in output_features:
            output_name = f[0]
            if output_name not in image_output_names:
                continue
            output_shape = f[1].dimensions
            if len(output_shape) == 2 or output_shape[0] == 1:
                is_grayscale = True
            elif output_shape[0] == 3:
                is_grayscale = False
            else:
                raise ValueError('Output must be RGB image or Grayscale')
            _set_deprocessing(
                is_grayscale,
                builder,
                deprocessing_args,
                mapping[output_name],
                output_name
            )

    if class_labels is not None:
        if isinstance(class_labels, Text):
            labels = [l.strip() for l in open(class_labels).readlines()]  # type: Sequence[Text]
        elif isinstance(class_labels, list):
            labels = class_labels
        else:
            raise TypeError(
                "synset variable of unknown type. Type found: {}. \
                Expected either string or list of strings."
                .format(type(class_labels),))

        builder.set_class_labels(
            class_labels=labels,
            predicted_feature_name=predicted_feature_name
        )

    def _add_informative_description(feature, raise_error=True):
        if feature.type.WhichOneof('Type') == 'multiArrayType':
            if feature.name in graph.onnx_coreml_shape_mapping and feature.name in graph.shape_dict:
                mapp = graph.onnx_coreml_shape_mapping[feature.name]
                onnx_shape = graph.shape_dict[feature.name]
                if raise_error: assert len(mapp) == len(onnx_shape), "Something wrong in shape"
                if len(mapp) == len(onnx_shape):
                    shape = []
                    for i in range(5):
                        if i in mapp:
                            shape += [int(onnx_shape[mapp.index(i)])]
                        else:
                            shape += [1]
                    msg = 'MultiArray of shape {}. The first and second dimensions correspond to sequence and batch size, respectively'.format(str(tuple(shape)))
                    feature.shortDescription += msg

    optional_input_names = []
    for tup in graph.optional_inputs:
        optional_input_names.append(tup[0])
    optional_output_names = []
    for tup in graph.optional_outputs:
        optional_output_names.append(tup[0])

    # add description for inputs and outputs shapes
    remove_input_id = []
    for i, input_ in enumerate(builder.spec.description.input):
        if input_.name not in optional_input_names:
            if not disable_coreml_rank5_mapping:
                _add_informative_description(input_)
        else:
            remove_input_id.append(i)
    remove_output_id = []
    for i, output_ in enumerate(builder.spec.description.output):
        if output_.name not in optional_output_names:
            if not disable_coreml_rank5_mapping:
                _add_informative_description(output_, raise_error=False)
        else:
            remove_output_id.append(i)

    for index in sorted(remove_input_id, reverse=True):
        del builder.spec.description.input[index]
    for index in sorted(remove_output_id, reverse=True):
        del builder.spec.description.output[index]


    if len(graph.optional_inputs) > 0 or len(graph.optional_outputs):
        builder.add_optionals(graph.optional_inputs, graph.optional_outputs)

    print("Translation to CoreML spec completed. Now compiling the CoreML model.")
    try:
        if DEBUG:
            import coremltools
            coremltools.models.utils.save_spec(builder.spec, '/tmp/node_model_raw_spec.mlmodel')
            from  coremltools.models.neural_network.printer import print_network_spec
            print_network_spec(builder.spec, style='coding')
        mlmodel = MLModel(builder.spec)
    except RuntimeError as e:
        raise ValueError('Compilation failed: {}'.format(str(e)))
    print('Model Compilation done.')


    # print information about all ops for which custom layers have been added
    if len(err.custom_layer_nodes) > 0:
        print('\n')
        print("Custom layers have been added to the CoreML model "
              "corresponding to the following ops in the onnx model: ")
        for i, node in enumerate(err.custom_layer_nodes):
            input_info = []
            for input_ in node.inputs:
                input_info.append((str(input_), graph.shape_dict.get(input_, str("Shape not available"))))
            output_info = []
            for output_ in node.outputs:
                output_info.append((str(output_), graph.shape_dict.get(output_, str("Shape not available"))))
            print("{}/{}: op type: {}, op input names and shapes: {}, op output names and shapes: {}".
                  format(i+1, len(err.custom_layer_nodes), node.op_type, str(input_info), str(output_info)))

    return mlmodel



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
# from __future__ import unicode_literals

import numpy as np
from typing import Any, Sequence, List
from onnx.backend.base import BackendRep, namedtupledict
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from coremltools.proto import FeatureTypes_pb2 as ft  #type: ignore
from coremltools.models import MLModel  #type: ignore
from typing import Dict, Any, Text, Tuple
from onnx import TensorProto
from ._graph import EdgeInfo


def _set_dtypes(input_dict, #type: Dict[Text, np._ArrayLike[Any]]
                model, #type: MLModel
                ):
    # type: (...) -> None
    spec = model.get_spec()
    for input_ in spec.description.input:
        if input_.type.HasField('multiArrayType') and input_.name in input_dict:
            if input_.type.multiArrayType.dataType == ft.ArrayFeatureType.INT32:
                input_dict[input_.name] = input_dict[input_.name].astype(np.int32)
            if input_.type.multiArrayType.dataType == ft.ArrayFeatureType.FLOAT32:
                input_dict[input_.name] = input_dict[input_.name].astype(np.float32)
            if input_.type.multiArrayType.dataType == ft.ArrayFeatureType.DOUBLE:
                input_dict[input_.name] = input_dict[input_.name].astype(np.float64)


class CoreMLRep(BackendRep):
    def __init__(self,
                 coreml_model,  # type: MLModel
                 onnx_outputs_info,  # type: Dict[Text, EdgeInfo]
                 useCPUOnly=False,  # type: bool
                 disable_rank5_mapping=False # type: bool
                 ):
        # type: (...) -> None
        super(CoreMLRep, self).__init__()
        self.model = coreml_model
        self.useCPUOnly = useCPUOnly
        self.disable_rank5_mapping = disable_rank5_mapping

        spec = coreml_model.get_spec()
        self.input_names = [str(i.name) for i in spec.description.input]
        self.output_names = [str(o.name) for o in spec.description.output]
        self.onnx_outputs_info = onnx_outputs_info  # type: Dict[Text, EdgeInfo]

    def run(self,
            inputs,  # type: Any
            **kwargs  # type: Any
            ):
        # type: (...) -> Tuple[Any, ...]
        super(CoreMLRep, self).run(inputs, **kwargs)
        inputs_ = inputs
        _reshaped = False
        if not self.disable_rank5_mapping:
            for i, input_ in enumerate(inputs_):
                shape = input_.shape
                if len(shape) == 4 or len(shape) == 2:
                    inputs_[i] = input_[np.newaxis, :]
                    _reshaped = True
                elif len(shape) == 3:
                    spec = self.model.get_spec()
                    spec_shape = [int(k) for k in spec.description.input[i].type.multiArrayType.shape]
                    prod = spec_shape[0] * spec_shape[1] * spec_shape[2]
                    onnx_shape = list(shape)
                    if onnx_shape != spec_shape:
                        if onnx_shape[2] == prod:
                            inputs_[i] = np.reshape(inputs_[i], [onnx_shape[0], onnx_shape[1]] + spec_shape)
                        elif onnx_shape[1] * onnx_shape[2] == prod:
                            inputs_[i] = np.reshape(inputs_[i], [1, onnx_shape[0]] + spec_shape)
        input_dict = dict(
            zip(self.input_names,
                map(np.array, inputs_)))
        _set_dtypes(input_dict, self.model) #type: ignore

        prediction = self.model.predict(input_dict, self.useCPUOnly)
        output_values = [prediction[name] for name in self.output_names]

        if not self.disable_rank5_mapping:
            for i, output_ in enumerate(output_values):
                shape = output_.shape
                #reshape the CoreML output to match Onnx's output shape
                try:
                    output_values[i] = np.reshape(output_, self.onnx_outputs_info[self.output_names[i]][2])  # type: ignore
                except RuntimeError:
                    print("Output '%s' shape incompatible between CoreML (%s) and onnx (%s)"
                        %(self.output_names[i], output_.shape,
                            self.onnx_outputs_info[self.output_names[i]]))
        
        ## Type Cast to ONNX expected output types
        for i, output_ in enumerate(output_values):
            output_type = self.onnx_outputs_info[self.output_names[i]][1]
            if TENSOR_TYPE_TO_NP_TYPE[output_type] != output_values[i].dtype:
                output_values[i] = output_values[i].astype(TENSOR_TYPE_TO_NP_TYPE[output_type])

        result = namedtupledict('Outputs',
                              self.output_names)(*output_values)  # type: Tuple[Any, ...]
        return result


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


import unittest
import onnx
from onnx_coreml import convert
import numpy as np
import numpy.testing as npt  # type: ignore
from tests._test_utils import _assert_outputs
import torch # type: ignore
import torch.nn as nn # type: ignore
import shutil
import tempfile
import os
import pytest

from coremltools.models.utils import macos_version

np.random.seed(10)
torch.manual_seed(10)

MIN_MACOS_VERSION_10_15 = (10, 15)

DEBUG = False

def _test_torch_model_single_io(torch_model, torch_input_shape, coreml_input_shape, disable_rank5_mapping=False):
    # run torch model
    torch_input = torch.rand(*torch_input_shape)
    torch_out_raw = torch_model(torch_input)
    if isinstance(torch_out_raw, tuple):
        torch_out = torch_out_raw[0].detach().numpy()
    else:
        torch_out = torch_out_raw.detach().numpy()

    # convert to onnx model
    model_dir = tempfile.mkdtemp()
    if DEBUG:
        model_dir = '/tmp'
    onnx_file = os.path.join(model_dir, 'torch_model.onnx')
    torch.onnx.export(torch_model, torch_input, onnx_file)
    onnx_model = onnx.load(onnx_file)

    # convert to coreml and run
    coreml_model = convert(onnx_model, disable_coreml_rank5_mapping=disable_rank5_mapping)

    output_name = [o.name for o in onnx_model.graph.output][0]
    initializer_names = {t.name for t in onnx_model.graph.initializer}
    input_name = [i.name for i in onnx_model.graph.input if i.name not in initializer_names][0]
    input_numpy = torch_input.detach().numpy()
    if disable_rank5_mapping:
        input_dict = {input_name: input_numpy} # type: ignore
    else:
        input_dict = {input_name: np.reshape(input_numpy, coreml_input_shape)}  # type: ignore
    coreml_out = coreml_model.predict(input_dict, useCPUOnly=True)[output_name]
    if DEBUG:
        coreml_model.save(model_dir + '/torch_model.mlmodel')
        print('coreml_out')
        print(np.squeeze(coreml_out))
        print('torch_out')
        print(np.squeeze(torch_out))
        print('coreml out shape ', coreml_out.shape)
        print('torch out shape: ', torch_out.shape)

    # compare
    _assert_outputs([torch_out], [coreml_out], decimal=4) # type: ignore

    # delete onnx model
    if not DEBUG:
        if os.path.exists(model_dir):
            shutil.rmtree(model_dir)

class OnnxModelTest(unittest.TestCase):

    def test_linear_no_bias(self, disable_rank5_mapping=False):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.simple_nn = nn.Sequential(nn.Linear(256, 128, bias=False), nn.ReLU())

            def forward(self, x):
                return self.simple_nn(x)

        torch_model = Net() # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1,256), (256), disable_rank5_mapping=disable_rank5_mapping) # type: ignore

    @unittest.skipIf(macos_version() < MIN_MACOS_VERSION_10_15,
                     'macOS 10.15+ required. Skipping test.')
    def test_linear_no_bias_disable_rank5_mapping(self):
        self.test_linear_no_bias(True)

    def test_linear_bias(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.simple_nn = nn.Sequential(nn.Linear(256, 128, bias=True), nn.ReLU())

            def forward(self, x):
                return self.simple_nn(x)

        torch_model = Net() # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1,256), (256)) # type: ignore

    def test_dynamic_reshape(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.conv = nn.Conv2d(in_channels=3,
                                      out_channels=32,
                                      kernel_size=(3, 3),
                                      stride=1, padding=0,
                                      bias=True)

            def forward(self, x):
                x = self.conv(x)
                x = x.view(x.size()[0], -1)
                return x

        torch_model = Net() # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 3, 100, 100), (3, 100, 100)) # type: ignore

    def test_const_initializer1(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.ones = torch.nn.Parameter(torch.ones(1,))

            def forward(self, x):
                y = x + self.ones
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 3), (3,))  # type: ignore


    def test_const_initializer2(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()

            def forward(self, x):
                y = x + torch.nn.Parameter(torch.ones(2, 3), requires_grad=False)
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 2, 3), (1, 2, 3))  # type: ignore

    def test_conv2D_transpose(self): # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.convT = torch.nn.ConvTranspose2d(1, 1, kernel_size=3, stride=2, output_padding=1, padding=1, groups=1)

            def forward(self, x):
                y = self.convT(x)
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 1, 2, 2), (1, 2, 2))  # type: ignore

    def test_conv2D_transpose_groups(self): # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.convT = torch.nn.ConvTranspose2d(4, 4, kernel_size=3, stride=2, output_padding=1, padding=1, groups=2)

            def forward(self, x):
                y = self.convT(x)
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 4, 8, 8), (4, 8, 8))  # type: ignore

    def test_conv2D_transpose_2(self): # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.convT = torch.nn.ConvTranspose2d(1, 1, kernel_size=3, stride=3, output_padding=2, padding=1, groups=1)

            def forward(self, x):
                y = self.convT(x)
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 1, 3, 3), (1, 3, 3))  # type: ignore

    def test_pow(self): # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()

            def forward(self, x):
                y = x.pow(3)
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (3, 2, 3), (3, 2, 3))  # type: ignore

    @unittest.skipIf(macos_version() < MIN_MACOS_VERSION_10_15,
                     'macOS 10.15+ required. Skipping test.')
    def test_lstm(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.lstm = nn.LSTM(input_size=256,
                                hidden_size=64,
                                num_layers=1)

            def forward(self, x):
                y = self.lstm(x)
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (3, 1, 256), (3, 1, 256), disable_rank5_mapping=True)  # type: ignore

    @unittest.skipIf(macos_version() < MIN_MACOS_VERSION_10_15,
                     'macOS 10.15+ required. Skipping test.')
    def test_bidirlstm(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.lstm = nn.LSTM(input_size=256,
                                hidden_size=64,
                                num_layers=1,
                                bidirectional=True)

            def forward(self, x):
                y = self.lstm(x)
                return y

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (3, 1, 256), (3, 1, 256), disable_rank5_mapping=True)  # type: ignore



    def test_1d_conv(self):
        class Net(nn.Module):
            def __init__(self, in_channels,
                               out_channels,
                               kernel_size,
                               stride=1,
                               dilation=1,
                               groups=1,
                               bias=True):

                super(Net, self).__init__()

                self.conv = torch.nn.Conv1d(in_channels,
                                            out_channels,
                                            kernel_size=kernel_size,
                                            stride=stride,
                                            padding=0,
                                            dilation=dilation,
                                            groups=groups,
                                            bias=bias)

                self.__padding = (kernel_size - 1) * dilation

            def forward(self, x):
                result = self.conv(x)
                if self.__padding != 0:
                    return result[:, :, :-self.__padding]
                return result

        B = 1
        Cin = 5
        Cout = 11
        k = 3
        Win = 15
        torch_model = Net(Cin, Cout, k)  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, Cin, Win), (Cin, 1, Win))  # type: ignore

    def test_conv1d_after_reshape(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.conv = torch.nn.Conv1d(in_channels=300,
                                            out_channels=32,
                                            kernel_size=3,
                                            stride=1,
                                            padding=0,
                                            bias=True)

            def forward(self, x):
                x = x.view(1, 300, 100)
                x = self.conv(x)
                return x

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 3, 100, 100), (3, 100, 100))  # type: ignore

    def test_bachnorm_after_reshape(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.conv = torch.nn.Conv1d(in_channels=300,
                                            out_channels=32,
                                            kernel_size=3,
                                            stride=1,
                                            padding=0,
                                            bias=True)
                self.bachnorm = nn.BatchNorm1d(32)

            def forward(self, x):
                x = x.view(1, 300, 100)
                x = self.conv(x)
                x = self.bachnorm(x)
                return x

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 3, 100, 100), (3, 100, 100))  # type: ignore

    def test_res_connect_downsampling_after_reshape(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.conv = torch.nn.Conv1d(in_channels=300,
                                            out_channels=32,
                                            kernel_size=3,
                                            stride=1,
                                            padding=1,
                                            bias=True)
                self.downsample = torch.nn.Conv1d(in_channels=300,
                                                  out_channels=32,
                                                  kernel_size=1,
                                                  stride=1,
                                                  padding=0,
                                                  bias=True)

            def forward(self, x):
                x = x.view(1, 300, 100)
                y = self.conv(x)
                res = self.downsample(x)
                return y + res

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 3, 100, 100), (3, 100, 100))  # type: ignore

    def test_fc_plus_convenet(self):  # type: () -> None
        class Net(nn.Module):
            def __init__(self, channel_size = 1, output_h = 16, output_w = 16, filter_num = 32, latent_size = 16):
                super(Net, self).__init__()
                self.channel_size = channel_size
                self.output_h = output_h
                self.output_w = output_w
                self.filter_num = filter_num
                self.latent_size = latent_size

                self.fc3 = nn.Linear(latent_size, 128)
                self.fc4 = nn.Linear(128, 256)

                self.relu = nn.ReLU()

                self.convt = nn.Sequential(
                    nn.ConvTranspose2d(256, self.filter_num * 4, 4, 1),
                    nn.BatchNorm2d(self.filter_num * 4),
                    nn.ReLU(inplace=True),
                    nn.ConvTranspose2d(self.filter_num * 4, self.filter_num * 2, 4, 1),
                    nn.BatchNorm2d(self.filter_num * 2),
                    nn.ReLU(inplace=True),
                    nn.ConvTranspose2d(self.filter_num * 2, self.filter_num, 4, 1),
                    nn.BatchNorm2d(self.filter_num),
                    nn.ReLU(inplace=True),
                    nn.ConvTranspose2d(self.filter_num, self.filter_num, 4, 1),
                    nn.BatchNorm2d(self.filter_num),
                    nn.ReLU(inplace=True),
                    nn.ConvTranspose2d(self.filter_num, 1, 4, 1),
                    nn.Sigmoid()
                )

            def forward(self, z):
                x = self.relu(self.fc3(z))
                deconv_input = self.fc4(x)
                deconv_input = deconv_input.view(-1, 256, 1, 1)
                x = self.convt(deconv_input)
                return x

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 16), (1, 1, 16))  # type: ignore

class ReshapeTransposeTests(unittest.TestCase):
    '''
    tests for models that have patterns like:
    rank(4) ---> reshape (rank 6) ----> transpose (rank 6) ----> reshape(4)
    '''

    @pytest.mark.xfail
    # is pytorch to onnx conversion correct?
    def test_pixel_shuffle_not_working(self):
        '''
        (1, c, h, w) --> reshape ---> (1, sh, sw, c/(sh*sw), h, w)
        --> transpose [0,1,4,2,5,3] ---> (1, sh, h, sw, w, c/(sh*sw))
        --> reshape ---> (1, c/(s1*s2), sh*h, sw*w)
        '''
        class Net(nn.Module):
            def __init__(self, upscale_factor=3):
                super(Net, self).__init__()
                self.upscale_factor = upscale_factor
                self.ps = nn.PixelShuffle(self.upscale_factor)

            def forward(self, x):
                return self.ps(x)

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 18, 4, 5), (18, 4, 5))  # type: ignore

    def test_pixel_shuffle_working(self):
        '''
        (1, c, h, w) --> reshape ---> (1, c/(sh*sw), sh, sw, h, w)
        --> transpose [0,1,4,2,5,3] ---> (1, sh, h, sw, w, c/(sh*sw))
        --> reshape ---> (1, c/(sh*sw), sh*h, sw*w)
        '''
        class Net(nn.Module):
            def __init__(self, C=12, H=4, W=6, sh=3, sw=2):
                super(Net, self).__init__()
                self.C = C
                self.H = H
                self.W = W
                self.sh = sh
                self.sw = sw

            def forward(self, x):
                y1 = x.view(1, self.C // (self.sh * self.sw), self.sh, self.sw, self.H, self.W).contiguous()
                y2 = y1.permute(0,1,4,2,5,3).contiguous()
                y3 = y2.view(1, self.C // (self.sh * self.sw), self.sh * self.H, self.sw * self.W).contiguous()
                return y3

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 12, 4, 6), (12, 4, 6))  # type: ignore

    def test_reorganize_1(self):
        '''
        (1, c, h, w) --> reshape ---> (1, c/(sh*sw), h, sh, w, sw)
        --> transpose [0,3,5,1,2,4] ---> (1, sh, sw, c/(sh*sw), h, w)
        --> reshape ---> (1, c*sh*sw, h/sh, w/sw)
        '''
        class Net(nn.Module):
            def __init__(self, C=12, H=4, W=6, sh=2, sw=3):
                super(Net, self).__init__()
                self.C = C
                self.H = H
                self.W = W
                self.sh = sh
                self.sw = sw

            def forward(self, x):
                y1 = x.view(1, self.C // (self.sh * self.sw), self.H, self.sh, self.W, self.sw).contiguous()
                y2 = y1.permute(0,3,5,1,2,4).contiguous()
                y3 = y2.view(1, self.C * (self.sh * self.sw), self.H // self.sh, self.W // self.sw).contiguous()
                return y3

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 12, 4, 6), (12, 4, 6))  # type: ignore

    def test_reorganize_2(self):
        '''
        (1, c, h, w) --> reshape ---> (1, c, h/sh, sh, w/sw, sw)
        --> transpose [0,1,2,4,3,5] ---> (1, c, h/sh, w/sw, sh, sw)
        --> reshape ---> (1, c*sh*sw, h/sh, w/sw)
        '''
        class Net(nn.Module):
            def __init__(self, C=12, H=4, W=6, sh=2, sw=3):
                super(Net, self).__init__()
                self.C = C
                self.H = H
                self.W = W
                self.sh = sh
                self.sw = sw

            def forward(self, x):
                y1 = x.view(1, self.C, self.H // self.sh, self.sh, self.W // self.sw, self.sw).contiguous()
                y2 = y1.transpose(4,3).contiguous()
                y3 = y2.view(1, self.C * (self.sh * self.sw), self.H // self.sh, self.W // self.sw).contiguous()
                return y3

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (1, 12, 4, 6), (12, 4, 6))  # type: ignore

class UnaryOperationTests(unittest.TestCase):
    '''
    Unary Operation Test cases
    '''
    ## Sqrt tests
    def sqrt_tensor(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return torch.sqrt(x)

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

class BinaryOperationTests(unittest.TestCase):
    '''
    Binary Operation Test cases
    '''
    ## Addition tests
    def add_same_shape(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return torch.add(x, y)

        y = torch.rand((18, 4, 5))
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    def add_same_shape_multiple(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return x + y + y1 + y2 + y3

        y = torch.rand((18, 4, 5))
        y1 = torch.rand((4, 5))
        y2 = torch.rand((18, 4, 5))
        y3 = 7.234
        
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    def add_tensor_scalar(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return torch.add(x, y)

        y = 5
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    def add_diff_shape(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return torch.add(x, y)

        y = torch.rand((4, 5))
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    ## Subtraction tests
    def sub_same_shape(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return torch.sub(x, y)

        y = torch.rand((18, 4, 5))
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    def sub_same_shape_multiple(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return x - y - y1 - y2 - y3

        y = torch.rand((18, 4, 5))
        y1 = torch.rand((4, 5))
        y2 = torch.rand((18, 4, 5))
        y3 = 7.234
        
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    def sub_tensor_scalar(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return torch.sub(x, y)

        y = 5
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    def sub_diff_shape(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return torch.sub(x, y)

        y = torch.rand((4, 5))
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

    def bianry_ops_mix_test(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return ((x * g + a) - d * (c + b) + (a * e - g) / e) / f

        a = torch.rand((18, 4, 5))
        b = torch.rand((4, 5))
        c = torch.rand((18, 4, 5))
        d = 7.234
        e = torch.rand((5))
        f = 8.234
        g = 5
         
        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 4, 5), disable_rank5_mapping=disable_rank5_mapping)  # type: ignore

class ReduceOperationTests(unittest.TestCase):
    '''
    Reduction Operation Test cases
    '''
    ## Reduction tests
    def reduce_sum(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return x.sum(dim=0)

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (4, 5), disable_rank5_mapping=disable_rank5_mapping)

    def reduce_mean(self, disable_rank5_mapping=True):
        class Net(nn.Module):
            def forward(self, x):
                return x.mean(dim=1) 

        torch_model = Net()  # type: ignore
        torch_model.train(False)
        _test_torch_model_single_io(torch_model, (18, 4, 5), (18, 5), disable_rank5_mapping=disable_rank5_mapping)

if __name__ == '__main__':
    unittest.main()
    #suite = unittest.TestSuite()
    #suite.addTest(OnnxModelTest("test_linear_no_bias_disable_rank5_mapping"))
    #unittest.TextTestRunner().run(suite)



from setuptools import setup, find_packages
from os import path
import sys


VERSION = '1.0b2'

here = path.abspath(path.dirname(__file__))

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    with open('README.md', "rb") as f:
        long_description = f.read().decode("UTF-8")


if sys.version_info[0] == 2:
    # Mypy doesn't work with Python 2
    mypy = []
elif sys.version_info[0] == 3:
    mypy = ['mypy==0.560']


setup(
    name='onnx-coreml',
    version=VERSION,
    packages=find_packages(exclude=['contrib', 'docs', 'test', 'example']),
    description='Convert ONNX (Open Neural Network Exchange)'
                'models into Apple CoreML format.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/onnx/onnx-coreml/',
    author='ONNX-CoreML Team',
    author_email='onnx-coreml@apple.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development'
    ],
    keywords='onnx coreml machinelearning ml coremltools converter neural',
    install_requires=[
        'click',
        'numpy',
        'onnx==1.5.0',
        'typing>=3.6.4',
        'typing-extensions>=3.6.2.1',
        'coremltools==3.0b3',
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest',
        'pytest-cov',
        'Pillow'
    ],
    extras_require={
        'mypy': mypy,
    },
    entry_points={
        'console_scripts': [
            'convert-onnx-to-coreml = onnx_coreml.bin.convert:onnx_to_coreml'
        ]
    },
)
